# -*- coding: utf-8 -*-
"""
Created on Sun Nov 16 12:09:43 2014

@author: edouard.duchesnay@gmail.com
"""

import numpy as np
import pandas as pd
import statsmodels.api as sm
from patsy import dmatrices
from collections import OrderedDict
from statsmodels.sandbox.stats.multicomp import multipletests

class MULM:
    """ Massive (application) of Univariate Linear Model on panda DataFrame.
    For all y in tagrets, for all x in regressors for all
    z in covar_models apply lm(y ~ x + z)

    Parameters
    ----------
    data: pandas DataFrame

    targets: list of strings, single columns of input dataframe,
        example ["y1", "y2"].

    regressors: list of string, single columns of input dataframe,
        example ["x1", "x2"].

    covar_models: list of string, formulae that apply on input dataframe,
        example ["z1", "z2", "z1+z2", "z1*z2"].

    out_filemane: string, output file of full statistics associate with the
    model. If None no such statistics are computed/stored. Warning if not
    None this slows dwn the execution speed.

    Return
    ------
    DataFrame with columns:
    "models", "target", "regressor", "covariate", "effect", "sd", "tvalue",
    "pvalue", "df".
    """

    def __init__(self, data, formulas, intercept=True):
        self.data = data
        self.formulas = formulas
        self.intercept = intercept
        #self.regressors = regressors
        #self.covar_models = covar_models
        #self.out_filemane = out_filemane

    def t_test(self, contrasts=None, out_filemane=None, anova=False):
        # Make sure contrasts is a list of list
        if contrasts is not None:
            if isinstance(contrasts, tuple):
                contrasts = list(contrasts)
            if not isinstance(contrasts, list):
                contrasts = [contrasts]
            def change(c):
                if isinstance(c, list): return c
                else: return [c]
            contrasts = [change(c) for c in contrasts]
        if out_filemane:
            out_fd = open(out_filemane,'w')
        #res_formulas, res_target, res_regressor, res_covariate = [], [], [], []
        res_formulas, res_target, res_contrasts = [], [], []
        res_effect, res_sd, res_tvalue, res_pvalue, res_df = [], [], [], [], []
        for formula in self.formulas:
            #formula = self.formulas[0]
            target, regressors = formula.split("~")
            # fit
            #dt = self.data[self.data[target].notnull()]
            y, X = dmatrices(formula, data=self.data, return_type='dataframe')
            if not self.intercept:
                X = X.ix[:, 1:]
            mod = sm.OLS(y, X)
            sm_fitted = mod.fit()
            if contrasts is not None and not anova:
                #contrasts = [1]
                if self.intercept:
                    contrasts_ = [[0] + c for c in contrasts]
                # pad with 0
                contrasts_ = [c + [0] * (X.shape[1] - len(c)) for c in contrasts_]
                #[0] + contrasts + [0] * (X.shape[1] - 2)
                contrasts_str = ["_".join(X.columns[np.where(c)[0]].tolist())
                    for c in contrasts_]
                # TODO FIXME IF CONTRASTS AS A MATRIX
                #contrasts_str = [X.columns[contrasts[i] != 0] for i in range(len(contrasts))]
            else:
                contrasts_ = np.identity(X.shape[1])
                contrasts_str = X.columns.tolist()
            #print contrasts_, formula, X.columns
            sm_ttest = sm_fitted.t_test(contrasts_)
            #
            res_formulas += [formula] * len(contrasts_)
            res_target += [target] * len(contrasts_)
            res_contrasts += contrasts_str
            #res_covariate += covars
            res_effect += sm_ttest.effect.ravel().tolist()
            res_sd += sm_ttest.sd.ravel().tolist()
            res_tvalue += sm_ttest.tvalue.ravel().tolist()
            res_pvalue += sm_ttest.pvalue.ravel().tolist()
            res_df += [sm_ttest.df_denom] * len(contrasts_)
            if out_filemane:
                out_fd.write("\n" + "=" * 78 + "\n")
                out_fd.write("== Model:" + formula + "\n")
                out_fd.write("=" * 78 + "\n")
                out_fd.write(sm_fitted.summary().as_text())
                out_fd.write("\n\n")
        if out_filemane:
            out_fd.close()
        o = OrderedDict()
        o["formula"] = res_formulas
        o["target"] = res_target
        o["contrast"] = res_contrasts
        o["effect"] = res_effect
        o["sd"] = res_sd
        o["tvalue"] = res_tvalue
        o["pvalue"] = res_pvalue
        o["df"] = res_df
        return pd.DataFrame(o)

    def t_test_maxT(self, contrasts, nperm=100):#:, alternative= "two_sided"):
        #alternatives = ["two_sided", "less", "greater", "one_sided_auto"]
        #if not alternative in alternatives:
        #    raise ValueError("Not a valid alternative")
        data_ori = self.data
        self.data = self.data.copy()
        stats = self.t_test(contrasts=contrasts, out_filemane=None)
        tmax = list()
        tmin = list()
        targets = set([formula.split("~")[0] for formula in self.formulas])
        for perm in range(nperm):
            print("Perm ", perm)
            # permut all targets
            for target in targets:
                self.data[target] = np.random.permutation(self.data[target])
            stats_p = self.t_test(contrasts=contrasts, out_filemane=None)
            tmax.append(np.max(stats_p.tvalue))
            tmin.append(np.min(stats_p.tvalue))
        self.data = data_ori
        self.tmax = np.array(tmax)
        self.tmin = np.array(tmin)
        self.tmax_2sided = np.max(np.abs(np.vstack([self.tmax, self.tmin])), axis=0)
        pvalues_twosided_maxT = np.asarray([np.sum(self.tmax_2sided >= np.abs(t)) / float(len(self.tmax))
                for t in stats.tvalue])
        pvalues_onesided_maxT = list()
        for t in stats.tvalue:
            if t >= 0:
                pvalues_onesided_maxT.append(np.sum(self.tmax >= t) / float(len(self.tmax)))
            else:
                pvalues_onesided_maxT.append(np.sum(self.tmin <= t) / float(len(self.tmin)))
        pvalues_onesided_maxT = np.asarray(pvalues_onesided_maxT)
        print("n pval two sided != one sided",
              np.sum(pvalues_onesided_maxT != pvalues_twosided_maxT))
        assert np.all(pvalues_onesided_maxT <= pvalues_twosided_maxT)
        stats["pvalues_onesided_maxT"] = pvalues_onesided_maxT
        stats["pvalues_twosided_maxT"] = pvalues_twosided_maxT
        stats["pvalue_fdr_bh"] = multipletests(stats.pvalue,
                                               method='fdr_bh')[1]
        return stats


if __name__ == "__main__":

    ###############
    # Build dataset
    ###############

    n = 100
    px_info = 5
    px_noise = 5
    regressors = ["x_%i" % i for i in range(px_info + px_noise)]

    pz = 3
    z_colnames = ["z_%i" % i for i in range(pz)]

    py_info = 2
    py_noise = 8
    targets = ["y_%i" % i for i in range(py_info + py_noise)]

    beta = np.concatenate([
        (np.arange(1, px_info+1) / float(px_info))[::-1],
        [0]*px_noise]).reshape((px_info + px_noise, 1))
    np.random.seed(1)
    X = np.random.randn(n, px_info + px_noise)
    Z = np.random.randn(n, pz)
    Y = np.random.randn(n, py_info + py_noise)
    # Causal model: add X on the first py_info variable
    Y[:, :py_info] += np.dot(X, beta) + np.dot(Z, [.1]*pz)[:, None] + 1.
    data = pd.DataFrame(np.concatenate([Y, X, Z], 1),
                        columns=targets + regressors + z_colnames)

    #######
    # Model
    #######
    covar_model = "+".join(z_colnames)
    formulas = ['%s~%s+%s' % (target, regressor, covar_model)
                for target in targets for regressor in regressors]
    model = MULM(data=data, formulas=formulas)

    #############################################
    # Statistics on regressors (of interest) only
    #############################################

    stats = model.t_test(contrasts=1, out_filemane=None)

    ntests = len(targets) *  len(regressors) * (1)
    assert stats.shape[0] == ntests
    # Check that P (Positive) P_expected-20% < P < P_expected-20%
    P = np.sum(stats.pvalue<0.05)
    TP = py_info * (px_info)
    ntests = len(targets) *  len(regressors) * (1) # no intercept
    P_expected = (ntests - TP) * .05 + TP
    #print P_expected, P
    assert (P_expected * 0.5 < P)  and (P < P_expected * 1.5)

    #########################################################"
    # Statistics on all regressors (of interest + covariates)
    #########################################################

    stats_full = model.t_test(out_filemane=None)
    # 1 + pz + 1 = regressor + covariates + intercept
    ntests = len(targets) *  len(regressors) * (1 + pz + 1)
    assert stats_full.shape[0] == ntests

    # Check that P (Positive) P_expected-20% < P < P_expected-20%
    P = np.sum((stats_full.pvalue<0.05) & (stats_full.contrast != 'Intercept'))
    TP = py_info * (px_info + pz)
    ntests = len(targets) *  len(regressors) * (1 + pz) # no intercept
    P_expected = (ntests - TP) * .05 + TP
    #print P_expected, P
    assert (P_expected * 0.5 < P)  and (P < P_expected * 1.5)


    ######################################
    # Tmax correction for mult comparisons
    ######################################
    model = MULM(data=data, formulas=formulas)
    self = model

    stats = model.t_test_maxT(contrasts=1, nperm=20)
    #stats = model.t_test_maxT(nperm=10, alternative="two_sided")
    # Check that P (Positive) P_expected-20% < P < P_expected-20%
    P = np.sum(stats.pvalues_onesided_maxT < 0.05)
    assert (P <= py_info * (px_info)) and (P > 1)

"""
run -i ~/git/pylearn-mulm/mulm/dataframe/mulm_dataframe.py


    def t_test2(self, full_model=False, out_filemane=None, anova=False):
        if out_filemane:
            out_fd = open(out_filemane,'w')
        res_formulas, res_target, res_regressor, res_covariate = [], [], [], []
        res_effect, res_sd, res_tvalue, res_pvalue, res_df = [], [], [], [], []
        for target in self.targets:
            for regressor in self.regressors:
                for covar_model in self.covar_models:
                    #print target, regressor, covar_model
                    if covar_model:
                        model = '%s~%s+%s' % (target, regressor, covar_model)
                    else:
                        model = '%s~%s' % (target, regressor)
                    # fit
                    dt = self.data[self.data[target].notnull()]
                    y, X = dmatrices(model, data=dt, return_type='dataframe')
                    mod = sm.OLS(y, X)
                    sm_fitted = mod.fit()
                    if not full_model and not anova:
                        contrasts = [[0, 1] + [0]*(X.shape[1]-2)]
                        covars = [covar_model]
                    else:
                        contrasts = np.identity(X.shape[1])
                        covars = X.columns.tolist()
                    sm_ttest = sm_fitted.t_test(contrasts)
                    #
                    res_formulas += [model] * len(contrasts)
                    res_target += [target] * len(contrasts)
                    res_regressor += [regressor] * len(contrasts)
                    res_covariate += covars
                    res_effect += sm_ttest.effect.ravel().tolist()
                    res_sd += sm_ttest.sd.ravel().tolist()
                    res_tvalue += sm_ttest.tvalue.ravel().tolist()
                    res_pvalue += sm_ttest.pvalue.ravel().tolist()
                    res_df += [sm_ttest.df_denom] * len(contrasts)
                    if out_filemane:
                        out_fd.write("\n" + "=" * 78 + "\n")
                        out_fd.write("== Model:" + model + "\n")
                        out_fd.write("=" * 78 + "\n")
                        out_fd.write(sm_fitted.summary().as_text())
                        out_fd.write("\n\n")
        if out_filemane:
            out_fd.close()
        o = OrderedDict()
        o["models"] = res_formulas
        o["target"] = res_target
        o["regressor"] = res_regressor
        o["covariate"] = res_covariate
        o["effect"] = res_effect
        o["sd"] = res_sd
        o["tvalue"] = res_tvalue
        o["pvalue"] = res_pvalue
        o["df"] = res_df
        return pd.DataFrame(o)
"""





"""
This module enable the identification of discriminative brain
connections when performing classification between two groups
with connectivity coefficients as features.

The classification is performed with a Support Vector Machine (SVM)
algorithm with a linear kernel. The C constant is set to 1.

References
----------
.. [1] Bernard Ng, Gaël Varoquaux, Jean-Baptiste Poline, Michael D. Greicius,
       Bertrand Thirion, "Transport on Riemannian Manifold for Connectivity-based
       brain decoding", IEEE Transactions on Medical Imaging, 2015.

Author: Dhaif BEKHA.


# TODO: code a function permutation_svc to estimate the null distribution without
# TODO: bootstrapping, because when sample size is too small, bootstrap with replacement fail.
"""

import numpy as np
from sklearn.svm import LinearSVC
from joblib import Parallel, delayed
import time
from scipy.stats import norm
from statsmodels.stats.multitest import multipletests
from sklearn.multiclass import OneVsRestClassifier
from conpagnon.utils.array_operation import array_rebuilder
from conpagnon.utils.folders_and_files_management import save_object
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from nilearn.plotting import plot_connectome
import os
from conpagnon.plotting.display import plot_matrix


def timer(start, end):
    """Print measured time between two point in the code

    Parameters
    ----------
    start: float
        The start of the measure
    end: float
        The end of the measure

    """
    hours, rem = divmod(end-start, 3600)
    minutes, seconds = divmod(rem, 60)
    print("{:0>2}:{:0>2}:{:05.2f}".format(int(hours), int(minutes), seconds))


def bootstrap_svc_(vectorized_connectivity_matrices, class_labels, bootstrap_number):
    """Fit Support Vector Machine with linear kernel on bootstrap sample
    """

    # Support Vector Machine with a linear kernel
    svc = LinearSVC()
    bootstrap_weights = []
    # indices of subjects number to bootstrap
    indices = np.arange(vectorized_connectivity_matrices.shape[0])
    for b in range(bootstrap_number):
        bootstrap_indices = np.random.choice(indices, size=vectorized_connectivity_matrices.shape[0],
                                             replace=True)
        bootstrap_matrices = vectorized_connectivity_matrices[bootstrap_indices, ...]
        bootstrap_class_label = class_labels[bootstrap_indices]
        # Fit SVC on bootstrap sample
        svc.fit(bootstrap_matrices, bootstrap_class_label)
        # Append classification weight for the current bootstrap
        bootstrap_weights.append(svc.coef_[0, ...])
    bootstrap_weights = np.array(bootstrap_weights)
    # Compute the normalized mean over bootstrap
    bootstrap_normalized_mean = (bootstrap_weights.mean(axis=0))/(bootstrap_weights.std(axis=0))

    return bootstrap_normalized_mean


def bootstrap_classification(features, class_labels, boot_indices, C=1):
    """Perform classification on two binary class for each sample generated
    by bootstrap (with replacement).

    Parameters
    ----------
    features: numpy.ndarray, shape (n_samples, n_features)
        The connectivity matrices in a vectorized form, that is
        each row is a subjects and each column is a pair of regions. Only
        the lower part of connectivity matrices should be given.
    class_labels: numpy.ndarray, shape (n_samples, )
        The class labels of each subjects, permuted one time.
    boot_indices: numpy.ndarray, shape (n_samples, )
        The array containing the indices of bootstrapped
        subjects.

    Returns
    -------
    output: nunmpy.ndarray, shape (n_features, )
        The weight of the linear SVM estimated on the boostrap sample.
    """

    svc = LinearSVC(C=C)

    bootstrap_matrices = features[boot_indices, ...]
    bootstrap_class_label = class_labels[boot_indices]
    # Fit SVC on bootstrap sample
    svc.fit(bootstrap_matrices, bootstrap_class_label)
    # Weight of features for the bootstrap sample
    bootstrap_coefficients = svc.coef_[0, ...]

    return bootstrap_coefficients


def bootstrap_svc(features, class_labels,
                  bootstrap_array_indices,
                  n_cpus_bootstrap=1,
                  verbose=0,
                  backend='multiprocessing',
                  C=1):
    """Perform classification between two binary class on
    bootstrapped samples.

    Parameters
    ----------
    features: numpy.ndarray, shape (n_samples, n_features)
        The connectivity matrices in a vectorized form, that is
        each row is a subjects and each column is a pair of regions. Only
        the lower part of connectivity matrices should be given.
    class_labels: numpy.ndarray, shape (n_sample, )
        The class labels of each subjects.
    bootstrap_array_indices: numpy.ndarray, shape (n_bootstrap, n_features)
        A array containing the bootstrapped indices. Each row
        contain the indices to generate a bootstrapped sample.
    n_cpus_bootstrap: int, optional
        The number CPU to be used concurrently during computation on
        bootstrap sample.  Default is one, like a classical
        for loop over bootstrap sample.
    backend: str, optional
        The method used to execute concurrent task. This argument
        is passed to the Parallel function in the joblib package.
        Default is multiprocessing.
    verbose: int, optional
        The verbosity level during parallel computation. This
        argument is passed to Parallel function in the joblib package.

    Returns
    -------
    output: numpy.ndarray, shape (n_bootstrap, n_features)
        The array of estimated features weights, for each
        bootstrapped sample.
    """
    bootstrap_number = bootstrap_array_indices.shape[0]
    results_bootstrap = Parallel(
        n_jobs=n_cpus_bootstrap, verbose=verbose,
        backend=backend)(delayed(bootstrap_classification)(
                            features=features,
                            class_labels=class_labels,
                            boot_indices=bootstrap_array_indices[b, ...],
                            C=C)
                         for b in range(bootstrap_number))

    return np.array(results_bootstrap)


def permutation_bootstrap_svc(features, class_labels_perm,
                              bootstrap_array_perm,
                              n_permutations=1000,
                              n_cpus_bootstrap=1,
                              backend='multiprocessing',
                              verbose_bootstrap=0,
                              verbose_permutation=0,
                              C=1):
    """Perform classification on two binary class for each sample generated
    by bootstrap (with replacement) and permuted class labels vector.

    Parameters
    ----------
    features: numpy.ndarray, shape (n_samples, n_features)
        The connectivity matrices in a vectorized form, that is
        each row is a subjects and each column is a pair of regions. Only
        the lower part of connectivity matrices should be given.
    class_labels_perm: numpy.ndarray, shape (n_permutations, n_samples)
        The class labels array: each row contain the subjects labels
        permuted one time.
    n_permutations: int, optional
        The number of permutations. Default is 1000.
    bootstrap_array_perm:,numpy.ndarray, shape (n_permutations, n_bootstrap, n_samples)
        A array which contain a number of bootstrap array indices for each permutations.
    n_cpus_bootstrap: int, optional
        The number CPU to be used concurrently during computation
        on bootstrap sample. Default is one, like a classical
        for loop over bootstrap sample.
    backend: str, optional
        The method used to execute concurrent task. This argument
        is passed to the Parallel function in the joblib package.
        Default is multiprocessing.
    verbose_bootstrap: int, optional
        The verbosity level during parallel computation. This
        argument is passed to Parallel function in the joblib package.
    verbose_permutation: int, optional
        If equal to 1, print the progression of the permutations testing.
        Default is 1.

    Returns
    -------
    output: numpy.ndarray, shape (n_features, )
        The normalized features weights mean, estimated by classification
        with a linear SVM, over bootstrap sample
    """

    # Initialization of the normalized mean features weight over
    # N permutations
    null_distribution = np.zeros((n_permutations, features.shape[1]))

    tic_permutations = time.time()
    for n in range(n_permutations):
        if verbose_permutation == 1:
            print('Performing permutation number {} out of {}'.format(n, n_permutations))
        # Generate a bootstrap array indices for the current permutations
        # Perform the classification of each bootstrap sample, but with the labels shuffled
        bootstrap_weight_perm = bootstrap_svc(features=features,
                                              class_labels=class_labels_perm[n, ...],
                                              bootstrap_array_indices=bootstrap_array_perm[n, ...],
                                              verbose=verbose_bootstrap,
                                              backend=backend,
                                              n_cpus_bootstrap=n_cpus_bootstrap,
                                              C=C)
        # Compute the normalized mean of weight for the current permutation
        normalized_mean_weight_perm = bootstrap_weight_perm.mean(axis=0) / bootstrap_weight_perm.std(axis=0)
        # Save it in the null distribution array
        null_distribution[n, ...] = normalized_mean_weight_perm
    tac_permutations = time.time() - tic_permutations
    print('Elapsed time for {} permutations: {} min'.format(n_permutations, tac_permutations/60))

    return null_distribution


def one_against_all_classification(features, class_labels, boot_indices):
    """Perform multi-class classification problem on a bootstrapped
    sample with a one versus all strategy.

    Parameters
    ----------
    features: numpy.ndarray, shape (n_samples, n_features)
        The connectivity matrices in a vectorized form, that is
        each row is a subjects and each column is a pair of regions. Only
        the lower part of connectivity matrices should be given.
    class_labels: numpy.ndarray, shape (n_samples, )
        The class labels of each subjects, permuted one time.
    boot_indices: numpy.ndarray, shape (n_samples, )
        The array containing the indices of bootstrapped
        subjects.

    Returns
    -------
    output: nunmpy.ndarray, shape (n_features, )
        The weight of the linear SVM estimated on the boostrap sample.
    """
    svc = LinearSVC()
    one_vs_all = OneVsRestClassifier(estimator=svc)

    bootstrap_matrices = features[boot_indices, ...]
    bootstrap_class_label = class_labels[boot_indices]
    # Fit SVC on bootstrap sample using a one class vs all classifier
    one_vs_all.fit(bootstrap_matrices, bootstrap_class_label)
    # Weight of features for the bootstrap sample for each classes
    bootstrap_coefficients = one_vs_all.coef_

    return bootstrap_coefficients


def one_against_all_bootstrap(features, class_labels,
                              bootstrap_array_indices,
                              n_cpus_bootstrap=1,
                              verbose=0,
                              backend='multiprocessing'):
    """Perform classification between two binary class on
    bootstrapped samples.

    Parameters
    ----------
    features: numpy.ndarray, shape (n_samples, n_features)
        The connectivity matrices in a vectorized form, that is
        each row is a subjects and each column is a pair of regions. Only
        the lower part of connectivity matrices should be given.
    class_labels: numpy.ndarray, shape (n_sample, )
        The class labels of each subjects.
    bootstrap_array_indices: numpy.ndarray, shape (n_bootstrap, n_features)
        A array containing the bootstrapped indices. Each row
        contain the indices to generate a bootstrapped sample.
    n_cpus_bootstrap: int, optional
        The number CPU to be used concurrently during computation on
        bootstrap sample.  Default is one, like a classical
        for loop over bootstrap sample.
    backend: str, optional
        The method used to execute concurrent task. This argument
        is passed to the Parallel function in the joblib package.
        Default is multiprocessing.
    verbose: int, optional
        The verbosity level during parallel computation. This
        argument is passed to Parallel function in the joblib package.

    Returns
    -------
    output: numpy.ndarray, shape (n_bootstrap, n_features)
        The array of estimated features weights, for each
        bootstrapped sample.
    """
    bootstrap_number = bootstrap_array_indices.shape[0]
    results_bootstrap = Parallel(
        n_jobs=n_cpus_bootstrap, verbose=verbose,
        backend=backend)(delayed(one_against_all_classification)(
                            features=features,
                            class_labels=class_labels,
                            boot_indices=bootstrap_array_indices[b, ...])
                         for b in range(bootstrap_number))

    return np.array(results_bootstrap)


def one_against_all_permutation_bootstrap(features, class_labels_perm,
                                          bootstrap_array_perm,
                                          n_classes,
                                          n_permutations=1000,
                                          n_cpus_bootstrap=1,
                                          backend='multiprocessing',
                                          verbose_bootstrap=0,
                                          verbose_permutation=0):
    """Perform classification on two binary class for each sample generated
    by bootstrap (with replacement) and permuted class labels vector.

    Parameters
    ----------
    features: numpy.ndarray, shape (n_samples, n_features)
        The connectivity matrices in a vectorized form, that is
        each row is a subjects and each column is a pair of regions. Only
        the lower part of connectivity matrices should be given.
    class_labels_perm: numpy.ndarray, shape (n_permutations, n_samples)
        The class labels array: each row contain the subjects labels
        permuted one time.
    n_permutations: int, optional
        The number of permutations. Default is 1000.
    bootstrap_array_perm:,numpy.ndarray, shape (n_permutations, n_bootstrap, n_samples)
        A array which contain a number of bootstrap array indices for each permutations.
    n_cpus_bootstrap: int, optional
        The number CPU to be used concurrently during computation
        on bootstrap sample. Default is one, like a classical
        for loop over bootstrap sample.
    backend: str, optional
        The method used to execute concurrent task. This argument
        is passed to the Parallel function in the joblib package.
        Default is multiprocessing.
    verbose_bootstrap: int, optional
        The verbosity level during parallel computation. This
        argument is passed to Parallel function in the joblib package.
    verbose_permutation: int, optional
        If equal to 1, print the progression of the permutations testing.
        Default is 1.

    Returns
    -------
    output: numpy.ndarray, shape (n_features, )
        The normalized features weights mean, estimated by classification
        with a linear SVM, over bootstrap sample
    """

    null_distribution = np.zeros((n_permutations, n_classes, features.shape[1]))

    tic_permutations = time.time()
    for n in range(n_permutations):
        if verbose_permutation == 1:
            print('Performing permutation number {} out of {}'.format(n, n_permutations))
        # Generate a bootstrap array indices for the current permutations
        # Perform the classification of each bootstrap sample, but with the labels shuffled
        bootstrap_weight_perm = one_against_all_bootstrap(features=features,
                                              class_labels=class_labels_perm[n, ...],
                                              bootstrap_array_indices=bootstrap_array_perm[n, ...],
                                              verbose=verbose_bootstrap,
                                              backend=backend,
                                              n_cpus_bootstrap=n_cpus_bootstrap)
        # Compute the normalized mean of weight for the current permutation
        normalized_mean_weight_perm = bootstrap_weight_perm.mean(axis=0) / bootstrap_weight_perm.std(axis=0)
        # Save it in the null distribution array
        null_distribution[n, ...] = normalized_mean_weight_perm
    tac_permutations = time.time() - tic_permutations
    print('Elapsed time for {} permutations: {} min'.format(n_permutations, tac_permutations/60))

    return null_distribution


def k_largest_index_argsort(arr, k, reverse_order=False):
    """Returns the k+1 largest element indices in a an array

    Parameters
    ----------
    arr: numpy.ndarray
        A multi-dimensional array.
    k: int
        The number of largest elements indices to return.
    reverse_order: bool
        If True, the indices are returned from the largest to
        smallest element.

    Returns
    -------
        output: numpy.ndarray
            The array of the k+1 largest element indices.
            The shape is the same of the input array.

    """
    idx = np.argsort(arr.ravel())[:-k - 1:-1]
    if reverse_order:
        idx = idx[::-1]
    return np.column_stack(np.unravel_index(idx, arr.shape))


def k_smallest_index_argsort(arr, k, reverse_order=False):
    """Returns the k+1 smallest element indices in a an array

    Parameters
    ----------
    arr: numpy.ndarray
        A multi-dimensional array.
    k: int
        The number of smallest elements indices to return.
    reverse_order: bool
        If True, the indices are returned from the largest to
        smallest element.

    Returns
    -------
        output: numpy.ndarray
            The array of the k+1 smallest element indices.
            The shape is the same of the input array.
    """

    idx = np.argsort(arr.ravel())[:k + 1:1]
    if reverse_order:
        idx = idx[::-1]
    return np.column_stack(np.unravel_index(idx, arr.shape))


def remove_reversed_duplicates(iterable):
    # Create a set for already seen elements
    seen = set()
    for item in iterable:
        # Lists are mutable so we need tuples for the set-operations.
        tup = tuple(item)
        if tup not in seen:
            # If the tuple is not in the set append it in REVERSED order.
            seen.add(tup[::-1])
            # If you also want to remove normal duplicates uncomment the next line
            # seen.add(tup)
            yield item


def rank_top_features_weight(coefficients_array, top_features_number,
                             features_labels):
    """Establish a ranking of the most important features from the classifier, based
    on weights magnitude.

    Parameters
    ----------
    coefficients_array: numpy.ndarray, shape (n_features, n_features)
        The 2D array containing the features weights to rank.
    top_features_number: int
        The desired number of top features.
    features_labels: list
        The features labels list

    Returns
    -------
    output 1: numpy.ndarray, shape (n_top_features + 1, )
        The desired top features weights
    output 2: numpy.ndarray, shape (n_top_features +1, 2)
        The positions of each top features weights
    output 3: numpy.ndarray, shape (n_top_features +1, 2)
        The roi labels couple corresponding to top features.
    """

    # Find the k largest/smallest index couple coefficients in the array
    top_positive_coefficients_indices = k_largest_index_argsort(arr=coefficients_array, k=top_features_number,
                                                                reverse_order=True)

    top_negative_coefficients_indices = k_smallest_index_argsort(arr=coefficients_array, k=top_features_number,
                                                                 reverse_order=False)
    # Stack the k largest/smallest couple coefficient vector array
    top_coefficients_indices = np.vstack([top_negative_coefficients_indices,
                                          top_positive_coefficients_indices])

    # In the top coefficient stack, we have symmetric pair, due to symmetric connectivity
    # behavior, we only keep one pairs
    top_coefficients_indices_ = np.array(list(remove_reversed_duplicates(top_coefficients_indices)))

    # We construct an array of feature name based on top_coefficients_ couple index
    features_labels_couple = np.array([features_labels[top_coefficients_indices_[i]]
                                       for i in range(top_coefficients_indices_.shape[0])])

    # Fetch the weight in the normalized mean array based on indices stacked
    # in top_coefficients_indices
    top_weights = coefficients_array[top_coefficients_indices_[:, 0], top_coefficients_indices_[:, 1]]

    return top_weights, top_coefficients_indices_, features_labels_couple


def features_weights_max_t_correction(null_distribution_features_weights,
                                      normalized_mean_weight):
    """Compute features weight corrected p values with
    the estimated null distribution of normalized mean features weight.

    Parameters
    ----------
    null_distribution_features_weights: numpy.ndarray, shape (n_permutations, (n_features*(n_features - 1)/2))
        The estimated null distribution of normalized mean of  features weights
        estimated with class labels permutations, and bootstrap.
    normalized_mean_weight: numpy.ndarray, shape ( (n_features*(n_features - 1)/2), )
        The normalized mean of features weight estimated on bootstrapped sample.

    Returns
    -------
    output 1: numpy.ndarray, shape (
    """
    # Retrieve the number of permutation and features
    n_permutations = null_distribution_features_weights.shape[0]
    n_features = null_distribution_features_weights.shape[1]

    null_min_and_max_distribution = np.zeros((n_permutations, 2))

    # Find minimum and maximum weight in the normalized mean for each permutations
    null_min_and_max_distribution[:, 0], null_min_and_max_distribution[:, 1] = \
        null_distribution_features_weights.min(axis=1), \
        null_distribution_features_weights.max(axis=1)

    # Compare each edges weight from the true mean weight normalized distribution to the minimum and
    # maximum null estimated distribution.

    # null distribution for maximum and minimum normalized weight
    sorted_null_maximum_dist = sorted(null_min_and_max_distribution[:, 1])
    sorted_null_minimum_dist = sorted(null_min_and_max_distribution[:, 0])

    # p values array
    p_values_max = np.zeros(n_features)
    p_values_min = np.zeros(n_features)

    for feature in range(normalized_mean_weight.shape[0]):
        p_values_max[feature] = \
            (len(np.where(sorted_null_maximum_dist > normalized_mean_weight[feature])[0]) / (n_permutations + 1))
        p_values_min[feature] = \
            (len(np.where(sorted_null_minimum_dist < normalized_mean_weight[feature])[0]) / (n_permutations + 1))

    return sorted_null_maximum_dist, sorted_null_minimum_dist, p_values_max, p_values_min


def features_weights_parametric_correction(null_distribution_features_weights,
                                           normalized_mean_weight,
                                           method='fdr_bh',
                                           alpha=.05):
    """Parametric estimation of p-values for each features weight using the estimated
    null distribution and fitting it's mean and standard deviation with normal law.

    Parameters
    ----------
    null_distribution_features_weights: numpy.ndarray of shape (n_permutation,  (n_features*(n_features - 1)/2))
        The normalized mean features weights for each permutations.
    normalized_mean_weight: numpy.ndarray, shape ( (n_features*(n_features - 1)/2), )
        The estimated normalized mean features weights from bootstrapped samples.
    method: str, optional
        The correction method. There are multiple possible choices, please
        consults the statsmodels library. Default is the False Discovery Rate
        correction (FDR).
    alpha: float, optional
        The type I error rate threshold. Default is 0.05.

    Returns
    -------
    output: numpy.ndarray, shape ((n_features*(n_features - 1)/2), )
        The corrected p-values array.

    """

    # Compute the p-values in a parametric way
    # The mean normalized features weight over all permutation
    mean_normalized_weight_perm = null_distribution_features_weights.mean(axis=0)
    # the standard deviation of the mean normalized weight over all permutation
    std_normalized_weight_perm = null_distribution_features_weights.std(axis=0)
    # p values array initialisation
    p_values_array = np.zeros(normalized_mean_weight.shape[0])
    # We compute for each features weight the corresponding
    # p values.
    for j in range(normalized_mean_weight.shape[0]):
        p_values_array[j] = 2*min(1-norm.cdf(x=normalized_mean_weight[j],
                                             loc=mean_normalized_weight_perm[j],
                                             scale=std_normalized_weight_perm[j]),
                                  norm.cdf(x=normalized_mean_weight[j],
                                           loc=mean_normalized_weight_perm[j],
                                           scale=std_normalized_weight_perm[j]))
    # Account with multiple comparison
    reject, p_values_corrected, _, _ = multipletests(pvals=p_values_array,
                                                     alpha=alpha,
                                                     method=method)

    return p_values_corrected


def find_top_features(normalized_mean_weight_array, labels_regions, top_features_number=50):
    """Find the top features weight in the normalized mean weight array, and mask the other
    features weight outside the ranking.

    Parameters
    ----------
    normalized_mean_weight_array: numpy.ndarray shape (n_features,n_features)
        The array of each normalized mean feature weight, computed after bootstrapping.
    labels_regions: list
        The list of feature label.
    top_features_number: int
        The top features number to keep.

    Returns
    -------
    output 1: numpy.ndarray, shape(n_features, n_features)
        The normalized mean weight array containing the top features weight values
        and zero elsewhere.
    output 2: numpy.ndarray, shape(top_features_number + 1, )
        The top features weights.
    output 3: numpy.ndarray, shape(top_features_number + 1, 2)
        The indices of the top features in the normalized mean weights array.
    output 4: numpy.ndarray, shape(top_features_number + 1, 2)
        The labels of the top features in the normalized mean weights array.
    """

    n_nodes = normalized_mean_weight_array.shape[0]
    # Find the top features among the normalized mean weight distribution

    top_weights, top_coefficients_indices, top_weight_labels = rank_top_features_weight(
        coefficients_array=normalized_mean_weight_array,
        top_features_number=top_features_number,
        features_labels=labels_regions)

    # Plot the top features weight on glass brain
    top_weights_mask = np.zeros((n_nodes, n_nodes), dtype=bool)
    top_weights_mask[top_coefficients_indices[:, 0], top_coefficients_indices[:, 1]] = True
    normalized_mean_weight_array_top_features = np.multiply(normalized_mean_weight_array,
                                                            top_weights_mask)
    normalized_mean_weight_array_top_features += normalized_mean_weight_array_top_features.T

    return normalized_mean_weight_array_top_features, top_weights, top_coefficients_indices, top_weight_labels


def find_significant_features_indices(p_positive_features_significant,
                                      p_negative_features_significant,
                                      features_labels):
    """Return regions indices and corresponding labels for s
    surviving features after permutation testing, for both
    negative and positive features weight.

    Parameters
    ----------
    p_positive_features_significant: numpy.ndarray, shape (n_features, n_features)
        An array containing the weight for a associated significant p-values features for
        positive weights, and zero elsewhere.
    p_negative_features_significant: numpy.ndarray, shape (n_features, n_features)
        An array containing the weight for a associated significant p-values features for
        negative weights, and zero elsewhere.
    features_labels: numpy.ndarray, shape (n_features, )
        The features labels.

    Returns
    -------
    output 1: numpy.ndarray, shape (n_significant_features, 2)
        The indices array of significant positive weighted features.
    output 2: numpy.ndarray, shape (n_significant_features, 2)
        The indices array of significant negative weighted features.
    output 3: numpy.ndarray, shape (n_significant_features, 2)
        The labels array of significant positive weighted features.
    output 4: numpy.ndarray, shape (n_significant_features, 2)
        The labels array of significant negative weighted features.

    """

    significant_positive_features_indices = \
        np.array(list(remove_reversed_duplicates(np.array(np.where(p_positive_features_significant != 0)).T)))
    significant_positive_features_labels = \
        np.array([features_labels[significant_positive_features_indices[i]] for i
                  in range(significant_positive_features_indices.shape[0])])

    significant_negative_features_indices = \
        np.array(list(remove_reversed_duplicates(np.array(np.where(p_negative_features_significant != 0)).T)))

    significant_negative_features_labels = \
        np.array([features_labels[significant_negative_features_indices[i]] for i
                  in range(significant_negative_features_indices.shape[0])])

    return significant_positive_features_indices, significant_negative_features_indices, \
           significant_positive_features_labels, significant_negative_features_labels


def compute_weight_distribution(vectorized_connectivity_matrices,
                                bootstrap_number,
                                n_permutations,
                                class_labels,
                                C=1,
                                n_cpus_bootstrap=1,
                                verbose_bootstrap=1,
                                verbose_permutations=1):

    # Number of subjects
    n_subjects = vectorized_connectivity_matrices.shape[0]

    # Indices to bootstrap
    indices = np.arange(n_subjects)
    # Generate a matrix containing all bootstrapped indices
    bootstrap_matrix = np.random.choice(a=indices, size=(bootstrap_number, n_subjects),
                                        replace=True)

    # Generate a permuted class labels array
    class_labels_permutation_matrix = np.array([np.random.permutation(class_labels)
                                                for n in range(n_permutations)])

    bootstrap_array_perm = np.random.choice(a=indices,
                                            size=(n_permutations, bootstrap_number,
                                                  n_subjects),
                                            replace=True)

    # Sanity check fo bootstrapped sample class labels for each permutation
    print('Check Bootstrapped class labels for each permutation...')
    for n in range(n_permutations):
        for b in range(bootstrap_number):
            bootstrapped_permuted_labels = class_labels_permutation_matrix[n, bootstrap_array_perm[n, b, ...]]
            count_labels_occurence = len(np.unique(bootstrapped_permuted_labels))
            if count_labels_occurence == 2:
                pass
            else:
                print('For the permutation # {}, the bootstrapped samples # {} is ill...'.format(n, b))
                print(b)
                # We replace the problematic bootstrap
                new_bootstrap_indices = np.random.choice(a=indices, size=len(indices),
                                                         replace=True)
                bootstrap_array_perm[n, b, ...] = new_bootstrap_indices
                new_bootstrap_class_labels_permuted = class_labels_permutation_matrix[n, new_bootstrap_indices]
                class_labels_permutation_matrix[n, ...] = new_bootstrap_class_labels_permuted

    print('Verifying that bootstrapped sample labels classes contain two labels....')
    for n in range(n_permutations):
        for b in range(bootstrap_number):
            bootstrapped_permuted_labels = class_labels_permutation_matrix[n, bootstrap_array_perm[n, b, ...]]
            count_labels_occurence = len(np.unique(bootstrapped_permuted_labels))
            if count_labels_occurence < 2:
                raise ValueError('Sample size seems too small to generate clean bootstrapped class labels \n '
                                 'with at least two classes !')
            else:
                pass
    print('Done checking bootstrapped class labels for each permutation.')

    # True weight obtain by bootstrapping
    print('Performing classification on {} bootstrap sample...'.format(bootstrap_number))
    bootstrap_weight = bootstrap_svc(features=vectorized_connectivity_matrices,
                                     class_labels=class_labels,
                                     bootstrap_array_indices=bootstrap_matrix,
                                     n_cpus_bootstrap=n_cpus_bootstrap,
                                     verbose=verbose_bootstrap,
                                     C=C)
    # Estimation of null distribution of normalized mean weight
    null_distribution = permutation_bootstrap_svc(features=vectorized_connectivity_matrices,
                                                  class_labels_perm=class_labels_permutation_matrix,
                                                  bootstrap_array_perm=bootstrap_array_perm,
                                                  n_permutations=n_permutations,
                                                  n_cpus_bootstrap=n_cpus_bootstrap,
                                                  verbose_bootstrap=verbose_bootstrap,
                                                  verbose_permutation=verbose_permutations,
                                                  C=C)

    normalized_mean_weight = bootstrap_weight.mean(axis=0) / bootstrap_weight.std(axis=0)

    return normalized_mean_weight, null_distribution


def discriminative_brain_connection_identification(vectorized_connectivity_matrices, class_labels,
                                                   class_names, save_directory,
                                                   n_permutations, bootstrap_number,
                                                   features_labels, features_colors,
                                                   n_nodes,
                                                   atlas_nodes,
                                                   first_class_mean_matrix,
                                                   second_class_mean_matrix,
                                                   top_features_number=100,
                                                   correction='fdr_bh',
                                                   alpha=0.05,
                                                   n_cpus_bootstrap=1,
                                                   verbose_bootstrap=1,
                                                   verbose_permutations=1,
                                                   write_report=True,
                                                   node_size=15,
                                                   C=1):
    """Identify important connection when performing a binary classification task.
    """
    # Number of subjects
    n_subjects = vectorized_connectivity_matrices.shape[0]

    # report name for features visualisation and parameters text report
    report_filename = 'features_identification_' + class_names[0] + '_' + class_names[1] + '_' + str(alpha) + \
                      '_' + correction + '.pdf'
    text_report_filename = 'features_identification_' + class_names[0] + '_' + class_names[1] + '_' + str(alpha) + \
                           '_' + correction + '.txt'

    normalized_mean_weight, null_distribution = compute_weight_distribution(
        vectorized_connectivity_matrices=vectorized_connectivity_matrices,
        bootstrap_number=bootstrap_number,
        n_permutations=n_permutations,
        class_labels=class_labels,
        C=C,
        n_cpus_bootstrap=n_cpus_bootstrap,
        verbose_bootstrap=verbose_bootstrap,
        verbose_permutations=verbose_permutations
    )
    save_object(object_to_save=normalized_mean_weight, saving_directory=save_directory,
                filename='normalized_mean_weight_' + class_names[0] + '_' + class_names[1] + '.pkl')

    # Save the null distribution to avoid
    save_object(object_to_save=null_distribution, saving_directory=save_directory,
                filename='null_distribution_' + class_names[0] + '_' + class_names[1] + '.pkl')

    # Rebuild a symmetric array from normalized mean weight vector
    normalized_mean_weight_array = array_rebuilder(normalized_mean_weight,
                                                   'numeric', diagonal=np.zeros(n_nodes))
    # Find top features
    normalized_mean_weight_array_top_features, top_weights, top_coefficients_indices, top_weight_labels = \
        find_top_features(normalized_mean_weight_array=normalized_mean_weight_array,
                          labels_regions=features_labels)

    labels_str = []
    for i in range(top_weight_labels.shape[0]):
        labels_str.append(str(top_weight_labels[i]))

    if correction == 'max_t':
        # Corrected p values with the maximum statistic
        sorted_null_maximum_dist, sorted_null_minimum_dist, p_value_positive_weights, p_value_negative_weights = \
            features_weights_max_t_correction(null_distribution_features_weights=null_distribution,
                                              normalized_mean_weight=normalized_mean_weight)

        # Rebuild vectorized p values array
        p_max_values_array = array_rebuilder(vectorized_array=p_value_positive_weights,
                                             array_type='numeric',
                                             diagonal=np.ones(n_nodes))

        p_min_values_array = array_rebuilder(vectorized_array=p_value_negative_weights,
                                             array_type='numeric',
                                             diagonal=np.ones(n_nodes))

        # Find p-values under the alpha threshold
        p_negative_features_significant = np.array(p_min_values_array < alpha, dtype=int)
        p_positive_features_significant = np.array(p_max_values_array < alpha, dtype=int)

        significant_positive_features_indices, significant_negative_features_indices, \
        significant_positive_features_labels, significant_negative_features_labels = find_significant_features_indices(
            p_positive_features_significant=p_positive_features_significant,
            p_negative_features_significant=p_negative_features_significant,
            features_labels=features_labels)

        # Take the mean difference and masking all connection except the surviving ones for
        # surviving negative and positive features weight
        mean_difference_positive_mask = np.multiply(first_class_mean_matrix - second_class_mean_matrix,
                                                    p_positive_features_significant)
        mean_difference_negative_mask = np.multiply(first_class_mean_matrix - second_class_mean_matrix,
                                                    p_negative_features_significant)

        # Mask for all significant connection, for both,
        # positive and negative weight
        p_all_significant_features = p_positive_features_significant + p_negative_features_significant

        # Take the mean difference for the overall significant features mask
        mean_difference_all_significant_features_mask = np.multiply(
            first_class_mean_matrix - second_class_mean_matrix,
            p_all_significant_features)

        if write_report is True:

            with PdfPages(os.path.join(save_directory, report_filename)) as pdf:
                # Plot the estimated null distribution
                plt.figure(constrained_layout=True)
                plt.hist(sorted_null_maximum_dist, 'auto', histtype='bar', alpha=0.5,
                         edgecolor='black')
                # The five 5% extreme values among maximum distribution
                p95 = np.percentile(sorted_null_maximum_dist, q=95)
                plt.axvline(x=p95, color='black')
                plt.title('Null distribution of maximum normalized weight mean')
                pdf.savefig()

                plt.figure()
                plt.hist(sorted_null_minimum_dist, 'auto', histtype='bar',
                         alpha=0.5, edgecolor='black')
                # The five 5% extreme values among minimum distribution
                p5 = np.percentile(sorted_null_minimum_dist, q=5)
                plt.axvline(x=p5, color='black')
                plt.title('Null distribution of minimum normalized weight mean')
                pdf.savefig()

                plt.figure()
                plot_connectome(adjacency_matrix=normalized_mean_weight_array,
                                node_coords=atlas_nodes, colorbar=True,
                                title='Features weight', node_size=node_size,
                                node_color=features_colors)
                pdf.savefig()

                # plot the top weight in a histogram fashion
                fig = plt.figure(figsize=(15, 10), constrained_layout=True)

                weight_colors = ['blue' if weight < 0 else 'red' for weight in top_weights]
                plt.bar(np.arange(len(top_weights)), list(top_weights),
                        color=weight_colors,
                        edgecolor='black',
                        alpha=0.5)
                plt.xticks(np.arange(0, len(top_weights)), labels_str,
                           rotation=60,
                           ha='right')
                for label in range(len(plt.gca().get_xticklabels())):
                    plt.gca().get_xticklabels()[label].set_color(weight_colors[label])
                plt.xlabel('Features names')
                plt.ylabel('Features weights')
                plt.title('Top {} features ranking of normalized mean weight'.format(top_features_number))
                pdf.savefig()

                plt.figure()
                plot_connectome(adjacency_matrix=normalized_mean_weight_array_top_features,
                                node_coords=atlas_nodes,
                                colorbar=True,
                                title='Top {} features weight'.format(top_features_number),
                                node_size=node_size,
                                node_color=features_colors)
                pdf.savefig()

                if np.where(p_all_significant_features == 1)[0].size != 0:
                    # Plot on glass brain the significant positive features weight
                    plt.figure()
                    plot_connectome(adjacency_matrix=p_positive_features_significant,
                                    node_coords=atlas_nodes, colorbar=True,
                                    title='Significant positive weight', edge_cmap='Reds',
                                    node_size=node_size,
                                    node_color=features_colors)
                    pdf.savefig()

                    # Plot on glass brain the mean difference in connectivity between
                    # the two groups for surviving positive weight
                    plt.figure()
                    plot_connectome(adjacency_matrix=mean_difference_positive_mask,
                                    node_coords=atlas_nodes, colorbar=True,
                                    title='Difference in connectivity {} - {} (positive weights)'.format(
                                        class_names[0],
                                        class_names[1]),
                                    node_size=node_size,
                                    node_color=features_colors)
                    pdf.savefig()

                    # Plot on glass brain the significant negative features weight
                    plt.figure()
                    plot_connectome(adjacency_matrix=p_negative_features_significant,
                                    node_coords=atlas_nodes, colorbar=True,
                                    title='Significant negative weight', edge_cmap='Blues',
                                    node_size=node_size,
                                    node_color=features_colors)
                    pdf.savefig()

                    plt.figure()
                    plot_connectome(adjacency_matrix=mean_difference_negative_mask,
                                    node_coords=atlas_nodes, colorbar=True,
                                    title='Difference in connectivity {} - {} (negative weights)'.format(
                                        class_names[0],
                                        class_names[1]),
                                    node_size=node_size,
                                    node_color=features_colors)
                    pdf.savefig()

                    plt.figure()
                    plot_connectome(adjacency_matrix=mean_difference_all_significant_features_mask,
                                    node_coords=atlas_nodes, colorbar=True,
                                    title='Difference in connectivity {} - {} (all weights)'.format(
                                        class_names[0],
                                        class_names[1]),
                                    node_size=node_size,
                                    node_color=features_colors)
                    pdf.savefig()

                    # Matrix view of significant positive and negative weight
                    plt.figure()
                    plot_matrix(matrix=p_negative_features_significant,
                                labels_colors='auto', mpart='all',
                                colormap='Blues', linecolor='black',
                                title='Significant negative weight',
                                vertical_labels=features_labels,
                                horizontal_labels=features_labels)
                    pdf.savefig()

                    plt.figure()
                    plot_matrix(matrix=p_positive_features_significant,
                                labels_colors='auto', mpart='all',
                                colormap='Reds', linecolor='black',
                                title='Significant positive weight',
                                vertical_labels=features_labels,
                                horizontal_labels=features_labels)
                    pdf.savefig()

                    plt.close("all")
        else:
            pass

    else:

        # Perform another type of correction like FDR, ....
        p_values_corrected = features_weights_parametric_correction(
            null_distribution_features_weights=null_distribution,
            normalized_mean_weight=normalized_mean_weight,
            method=correction)

        p_values_corrected_array = array_rebuilder(vectorized_array=p_values_corrected,
                                                   array_type='numeric',
                                                   diagonal=np.ones(n_nodes))

        # Find p values under alpha threshold for negative and positive weight features
        p_negative_features_significant = np.array(
            (p_values_corrected_array < alpha) & (normalized_mean_weight_array < 0),
            dtype=int)
        p_positive_features_significant = np.array(
            (p_values_corrected_array < alpha) & (normalized_mean_weight_array > 0),
            dtype=int)

        significant_positive_features_indices, significant_negative_features_indices, \
        significant_positive_features_labels, significant_negative_features_labels = find_significant_features_indices(
            p_positive_features_significant=p_positive_features_significant,
            p_negative_features_significant=p_negative_features_significant,
            features_labels=features_labels)

        # Take the mean difference and masking all connection except the surviving ones for
        # surviving negative and positive features weight
        mean_difference_positive_mask = np.multiply(first_class_mean_matrix - second_class_mean_matrix,
                                                    p_positive_features_significant)
        mean_difference_negative_mask = np.multiply(first_class_mean_matrix - second_class_mean_matrix,
                                                    p_negative_features_significant)

        # Mask for all significant connection, for both,
        # positive and negative weight
        p_all_significant_features = p_positive_features_significant + p_negative_features_significant

        # Take the mean difference for the overall significant features mask
        mean_difference_all_significant_features_mask = np.multiply(
            first_class_mean_matrix - second_class_mean_matrix,
            p_all_significant_features)
        if write_report is True:

            # plot the top weight in a histogram fashion
            with PdfPages(os.path.join(save_directory, report_filename)) as pdf:

                plt.figure()
                plot_connectome(adjacency_matrix=normalized_mean_weight_array,
                                node_coords=atlas_nodes, colorbar=True,
                                title='Features weight',
                                node_size=node_size,
                                node_color=features_colors)
                pdf.savefig()

                fig = plt.figure(figsize=(15, 10), constrained_layout=True)
                weight_colors = ['blue' if weight < 0 else 'red' for weight in top_weights]
                plt.bar(np.arange(len(top_weights)), list(top_weights),
                        color=weight_colors,
                        edgecolor='black',
                        alpha=0.5)
                plt.xticks(np.arange(0, len(top_weights)), labels_str,
                           rotation=60,
                           ha='right')
                for label in range(len(plt.gca().get_xticklabels())):
                    plt.gca().get_xticklabels()[label].set_color(weight_colors[label])
                plt.xlabel('Features names')
                plt.ylabel('Features weights')
                plt.title('Top {} features ranking of normalized mean weight'.format(top_features_number))
                pdf.savefig()

                plt.figure()
                plot_connectome(adjacency_matrix=normalized_mean_weight_array_top_features,
                                node_coords=atlas_nodes, colorbar=True,
                                title='Top {} features weight'.format(top_features_number),
                                node_size=node_size,
                                node_color=features_colors)
                pdf.savefig()
                if np.where(p_all_significant_features == 1)[0].size != 0:
                    # Plot on glass brain the significant positive features weight
                    plt.figure()
                    plot_connectome(adjacency_matrix=p_positive_features_significant,
                                    node_coords=atlas_nodes, colorbar=True,
                                    title='Significant positive weight after {} correction'.format(correction),
                                    edge_cmap='Reds',
                                    node_size=node_size,
                                    node_color=features_colors)

                    pdf.savefig()

                    # the two groups for surviving positive weight
                    plt.figure()
                    plot_connectome(adjacency_matrix=mean_difference_positive_mask,
                                    node_coords=atlas_nodes, colorbar=True,
                                    title='Difference in connectivity {} - {} (positive weights)'.format(
                                        class_names[0],
                                        class_names[1]),
                                    node_size=node_size,
                                    node_color=features_colors)
                    pdf.savefig()

                    # Plot on glass brain the significant negative features weight
                    plt.figure()
                    plot_connectome(adjacency_matrix=p_negative_features_significant,
                                    node_coords=atlas_nodes, colorbar=True,
                                    title='Significant negative weight after {} correction'.format(correction),
                                    edge_cmap='Blues',
                                    node_size=node_size,
                                    node_color=features_colors)
                    pdf.savefig()

                    plt.figure()
                    plot_connectome(adjacency_matrix=mean_difference_negative_mask,
                                    node_coords=atlas_nodes, colorbar=True,
                                    title='Mean difference in connectivity {} - {} (negative weights)'.format(
                                        class_names[0],
                                        class_names[1]),
                                    node_size=node_size,
                                    node_color=features_colors)
                    pdf.savefig()

                    plt.figure()
                    plot_connectome(adjacency_matrix=mean_difference_all_significant_features_mask,
                                    node_coords=atlas_nodes, colorbar=True,
                                    title='Difference in connectivity {} - {} (all weights)'.format(
                                        class_names[0],
                                        class_names[1]),
                                    node_size=node_size,
                                    node_color=features_colors)
                    pdf.savefig()

                    # Matrix view of significant positive and negative weight
                    plt.figure()
                    plot_matrix(matrix=p_negative_features_significant,
                                labels_colors=features_colors, mpart='all',
                                colormap='Blues', linecolor='black',
                                title='Significant negative weight after {} correction'.format(correction),
                                vertical_labels=features_labels, horizontal_labels=features_labels)
                    pdf.savefig()

                    plt.figure()
                    plot_matrix(matrix=p_positive_features_significant,
                                labels_colors=features_colors,
                                mpart='all',
                                colormap='Reds', linecolor='black',
                                title='Significant positive weight after {} correction'.format(correction),
                                vertical_labels=features_labels, horizontal_labels=features_labels)
                    pdf.savefig()

                    plt.close("all")
        else:
            pass

    # Write a small report in a text file
    with open(os.path.join(save_directory, text_report_filename), 'w') as output_results:
        # Write parameters
        output_results.write('------------ Parameters ------------')
        output_results.write('\n')
        output_results.write('Number of subjects: {}'.format(n_subjects))
        output_results.write('\n')
        output_results.write('Groups: {}'.format(class_names))
        output_results.write('\n')
        output_results.write('Groups labels: {}'.format(class_labels))
        output_results.write('\n')
        output_results.write('Bootstrap number: {}'.format(bootstrap_number))
        output_results.write('\n')
        output_results.write('Number of permutations: {}'.format(n_permutations))
        output_results.write('\n')
        output_results.write('Alpha threshold: {}'.format(alpha))
        output_results.write('\n')
        output_results.write('Regularization parameters C: {}'.format(C))
        output_results.write('\n')
        output_results.write('\n')
        output_results.write('------------ Results ------------')
        output_results.write('\n')
        output_results.write('------------ Discriminative connections for negative features weight ------------')
        output_results.write('\n')
        # Write the labels of features with negative weight identified
        for negative_feature in range(len(significant_negative_features_labels)):
            output_results.write('\n')
            output_results.write('{} <-> {}, (indices: {} <-> {})'.format(
                significant_negative_features_labels[negative_feature][0],
                significant_negative_features_labels[negative_feature][1],
                significant_negative_features_indices[negative_feature][0],
                significant_negative_features_indices[negative_feature][1]))

        output_results.write('\n')
        output_results.write('\n')
        output_results.write('------------ Discriminative connections for positive features weight ------------')
        output_results.write('\n')
        # Write the labels of features with negative weight identified
        for positive_feature in range(len(significant_positive_features_labels)):
            output_results.write('\n')
            output_results.write('{} <-> {}, (indices: {} <-> {})'.format(
                significant_positive_features_labels[positive_feature][0],
                significant_positive_features_labels[positive_feature][1],
                significant_positive_features_indices[positive_feature][0],
                significant_positive_features_indices[positive_feature][1]))

    return normalized_mean_weight, null_distribution, p_all_significant_features














# -*- coding: utf-8 -*-
"""
Created on Tue Jun 25 13:25:41 2013

@author: ed203246
"""
import numpy as np
import scipy
from sklearn.preprocessing import scale
from scipy import stats

class MUPairwiseCorr:
    """Mass-univariate pairwise correlations. Given two arrays X [n_samples x p]
    and Y [n_samples x q]. Fit p x q independent linear models. Prediction
    and stats return [p x q] array.


    Example
    -------
    >>> import numpy as np
    >>> from mulm import MUPairwiseCorr
    >>> X = np.random.randn(10, 5)
    >>> Y = np.random.randn(10, 3)
    >>> corr = MUPairwiseCorr()
    >>> corr.fit(X, Y)
    <mulm.models.MUPairwiseCorr instance at 0x30da878>
    >>> f, p = corr.stats_f(X, Y)
    >>> print f.shape
    (5, 3)
    """
    def __init__(self, **kwargs):
        pass

    def fit(self, X, Y):
        Xs = scale(X, copy=True)
        Ys = scale(Y, copy=True)
        self.n_samples = X.shape[0]
        self.Corr_ = np.dot(Xs.T, Ys) / self.n_samples
        return self

    def predict(self, X):
        pass

    def stats_f(self, pval=True):
        R2 = self.Corr_ ** 2
        df_res = self.n_samples - 2
        f_stats = R2 * df_res / (1 - R2)
        if not pval:
            return (f_stats, None)
        else:
            p_vals = stats.f.sf(f_stats, 1, df_res)
            return f_stats, p_vals

class MUOLS:
    """Mass-univariate linear modeling based Ordinary Least Squares.
    Given two arrays X (n_samples, p) and Y (n_samples, q).
    Fit q independent linear models, ie., for all y in Y fit: lm(y ~ X)

    Example
    -------
    """

    def _block_slices(self, dim_size, block_size):
        """Generator that yields slice objects for indexing into
        sequential blocks of an array along a particular axis
        """
        count = 0
        while True:
            yield slice(count, count + block_size, 1)
            count += block_size
            if count >= dim_size:
                raise StopIteration

    def __init__(self, Y, X):
        self.coef = None
        if X.shape[0] != Y.shape[0]:
            raise ValueError('matrices are not aligned')
        self.X = X  # TODO PERFORM BASIC CHECK ARRAY
        self.Y = Y  # TODO PERFORM BASIC CHECK ARRAY

    def fit(self, block=False, max_elements=2 ** 27):
        """Use block=True for huge matrices Y.
        Operations block by block to optimize time and memory.
        max_elements: block dimension (2**27 corresponds to 1Go)
        """
        self.block = block
        self.max_elements = max_elements
        self.pinv = scipy.linalg.pinv2(self.X)
        n, p = self.Y.shape
        q = self.X.shape[1]
        if self.block:
            if self.max_elements < n:
                raise ValueError('the maximum number of elements is too small')
            max_cols = int(self.max_elements / n)
        else:
            max_cols = p
        self.coef = np.zeros((q, p))
        self.err_ss = np.zeros(p)
        for pp in self._block_slices(p, max_cols):
            if isinstance(self.Y, np.memmap):
                Y_block = self.Y[:, pp].copy()  # copy to force a read
            else: Y_block = self.Y[:, pp]
            #Y_block = self.Y[:, pp]
            self.coef[:, pp] = np.dot(self.pinv, Y_block)
            y_hat = np.dot(self.X, self.coef[:, pp])
            err = Y_block - y_hat
            del Y_block, y_hat
            self.err_ss[pp] = np.sum(err ** 2, axis=0)
            del err

#        self.coef = np.dot(self.pinv, self.Y)
#        y_hat = self.predict(self.X)
#        err = self.Y - y_hat
#        self.err_ss = np.sum(err ** 2, axis=0)
        return self

    def predict(self, X):
        #from sklearn.utils import safe_asarray
        import numpy as np
        #X = safe_asarray(X) # TODO PERFORM BASIC CHECK ARRAY
        pred_y = np.dot(X, self.coef)
        return pred_y


    def t_test(self, contrasts, pval=False, two_tailed=True):
        """Compute statistics (t-scores and p-value associated to contrast)

        Parameters
        ----------
        contrasts: The k contrasts to be tested, some list or array
        that can be casted to an k x p array.

        pval: boolean
            compute pvalues (default is false)

        two_tailed: boolean
            one-tailed test or a two-tailed test (default True)

        Return
        ------
        tstats (k, p) array, pvals (k, p) array, df (k,) array

        Example
        -------
        >>> import numpy as np
        >>> import mulm
        >>> X = np.random.randn(100, 5)
        >>> Y = np.random.randn(100, 10)
        >>> beta = np.random.randn(5, 1)
        >>> Y[:, :2] += np.dot(X, beta)
        >>> contrasts = np.identity(X.shape[1])
        >>> mod = mulm.MUOLS(Y, X).fit()
        >>> tvals, pvals, df = mod.t_test(contrasts, pval=True, two_tailed=True)
        """
        contrasts = np.atleast_2d(np.asarray(contrasts))
        n = self.X.shape[0]
        t_stats_ = list()
        p_vals_ = list()
        df_ = list()
        for contrast in contrasts:
            #ccontrasts = np.asarray(contrasts)
            # t = c'beta / std(c'beta)
            # std(c'beta) = sqrt(var_err (c'X+)(X+'c))
            #Xpinv = scipy.linalg.pinv(X)
            cXpinv = np.dot(contrast, self.pinv)
            R = np.eye(n) - np.dot(self.X, self.pinv)
            df = np.trace(R)
            ## Broadcast over ss errors
            var_errors = self.err_ss / df
            std_cbeta = np.sqrt(var_errors * np.dot(cXpinv, cXpinv.T))
            t_stats = np.dot(contrast, self.coef) / std_cbeta
            p_vals = None
            if pval is not None:
                if two_tailed:
                    p_vals = stats.t.sf(np.abs(t_stats), df) * 2
                else:
                    p_vals = stats.t.sf(t_stats, df)
            t_stats_.append(t_stats)
            p_vals_.append(p_vals)
            df_.append(df)
        return np.asarray(t_stats_), np.asarray(p_vals_), np.asarray(df_)

    def t_test_maxT(self, contrasts, nperms=1000, two_tailed=True, **kwargs):
        """Correct for multiple comparisons using maxT procedure. See t_test()
        For all parameters.

        Example
        -------
        >>> import numpy as np
        >>> import mulm
        >>> import pylab as plt
        >>> n = 100
        >>> px = 5
        >>> py_info = 2
        >>> py_noize = 100
        >>> beta = np.array([1, 0, .5] + [0] * (px - 4) + [2]).reshape((px, 1))
        >>> X = np.hstack([np.random.randn(n, px-1), np.ones((n, 1))]) # X with intercept
        >>> Y = np.random.randn(n, py_info + py_noize)
        >>> Y[:, :py_info] += np.dot(X, beta)
        >>> contrasts = np.identity(X.shape[1])
        >>> mod = mulm.MUOLS(Y, X)
        >>> tvals, maxT, df = mod.t_test_maxT(contrasts, two_tailed=True)
        """
        #contrast = [0, 1] + [0] * (X.shape[1] - 2)
        tvals, _, df = self.t_test(contrasts=contrasts, pval=False, **kwargs)
        max_t = list()
        for i in range(nperms):
            
            perm_idx = np.random.permutation(self.X.shape[0])
            Xp = self.X[perm_idx, :]
            muols = MUOLS(self.Y, Xp).fit(block=self.block,
                                          max_elements=self.max_elements)
            tvals_perm, _, _ = muols.t_test(contrasts=contrasts, pval=False,
                                            two_tailed=two_tailed)
            if two_tailed:
                tvals_perm = np.abs(tvals_perm)
            max_t.append(np.max(tvals_perm, axis=1))
            del muols
        max_t = np.array(max_t)
        tvals_ = np.abs(tvals) if two_tailed else tvals
        pvalues = np.array(
            [np.array([np.sum(max_t[:, con] >= t) for t in tvals_[con, :]])\
                / float(nperms) for con in range(contrasts.shape[0])])
        return tvals, pvalues, df, max_t

    def t_test_minP(self, contrasts, nperms=10000, two_tailed=True, **kwargs):
        """Correct for multiple comparisons using minP procedure.
        For all parameters.

        Example
        -------
        >>> import numpy as np
        >>> import mulm
        >>> import pylab as plt
        >>> n = 100
        >>> px = 5
        >>> py_info = 2
        >>> py_noize = 100
        >>> beta = np.array([1, 0, .5] + [0] * (px - 4) + [2]).reshape((px, 1))
        >>> X = np.hstack([np.random.randn(n, px-1), np.ones((n, 1))]) # X with intercept
        >>> Y = np.random.randn(n, py_info + py_noize)
        >>> Y[:, :py_info] += np.dot(X, beta)
        >>> contrasts = np.identity(X.shape[1])
        >>> tvals, maxT, df = mod.t_test_minP(contrasts, two_tailed=True)
        """
        tvals, pvals, df = self.t_test(contrasts=contrasts, pval=True, **kwargs)
        min_p = np.ones((contrasts.shape[0], nperms))
        perm_idx = np.zeros((self.X.shape[0], nperms + 1), dtype='int')
        for i in range(self.Y.shape[1]):
            Y_curr = self.Y[:, i]
            Yp_curr = np.zeros((self.X.shape[0], nperms + 1))

            for j in range(nperms + 1):
                if i == 0:
                    perm_idx[:, j] = np.random.permutation(self.X.shape[0])
                Yp_curr[:, j] = Y_curr[perm_idx[:, j]]
            muols = MUOLS(Yp_curr, self.X).fit()
            tvals_perm, _, _ = muols.t_test(contrasts=contrasts, pval=False,
                                            two_tailed=two_tailed)
            if two_tailed:
                tvals_perm = np.abs(tvals_perm)
            pval_perm = np.array(
               [np.array([((np.sum(tvals_perm[con, :] >= tvals_perm[con, k])) - 1) \
                         for k in range(nperms)]) / float(nperms) \
                             for con in range(contrasts.shape[0])])
            min_p = np.array(
               [(np.min(np.vstack((min_p[con, :], pval_perm[con, :])), axis=0)) \
                         for con in range(contrasts.shape[0])])
        pvalues = np.array(
               [np.array([np.sum(min_p[con, :] <= p) \
                         for p in pvals[con, :]]) / float(nperms) \
                             for con in range(contrasts.shape[0])])
        return tvals, pvalues, df

    def f_test(self, contrast, pval=False):
        from sklearn.utils import array2d
        #Ypred = self.predict(self.X)
        #betas = self.coef
        #ss_errors = np.sum((self.Y - self.y_hat) ** 2, axis=0)
        C1 = array2d(contrast).T
        n, p = self.X.shape
        #Xpinv = scipy.linalg.pinv(X)
        rank_x = np.linalg.matrix_rank(self.pinv)
        C0 = np.eye(p) - np.dot(C1, scipy.linalg.pinv2(C1))  # Ortho. cont. to C1
        X0 = np.dot(self.X, C0)  # Design matrix of the reduced model
        X0pinv = scipy.linalg.pinv2(X0)
        rank_x0 = np.linalg.matrix_rank(X0pinv)
        # Find the subspace (X1) of Xc1, which is orthogonal to X0
        # The projection matrix M due to X1 can be derived from the residual
        # forming matrix of the reduced model X0
        # R0 is the residual forming matrix of the reduced model
        R0 = np.eye(n) - np.dot(X0, X0pinv)
        # R is the residual forming matrix of the full model
        R = np.eye(n) - np.dot(self.X, self.pinv)
        # compute the projection matrix
        M = R0 - R
        #Ypred = np.dot(self.X, betas)
        y_hat = self.predict(self.X)
        SS = np.sum(y_hat * np.dot(M, y_hat), axis=0)
        df_c1 = rank_x - rank_x0
        df_res = n - rank_x
        ## Broadcast over self.err_ss of Y
        f_stats = (SS * df_res) / (self.err_ss * df_c1)
        if not pval:
            return (f_stats, None)
        else:
            p_vals = stats.f.sf(f_stats, df_c1, df_res)
            return f_stats, p_vals

    def stats_f_coefficients(self, X, Y, contrast, pval=False):
        return self.stats_f(contrast, pval=pval)




# -*- coding: utf-8 -*-
"""
Created on Mon Jul 28 15:39:24 2014

@author:  edouard.duchesnay@gmail.com
@license: BSD-3-Clause

python ~/git/datamind/descriptive/descriptive_statistics.py -i /tmp/db.csv -o /tmp/db.xls -e "ID"
"""
import pandas as pd
import argparse, sys

def describe_df_basic(data):
    basic_desc = pd.DataFrame([[x, str(data[x].dtype), int(str(data[x].dtype)!="object"),
                          pd.notnull(data[x]).sum()] for x in data.columns],
                          columns=["variable", "type", "isnumeric", "count"])
    return basic_desc

def describe_df(data, exclude_from_cat_desc=[]):
    basic = describe_df_basic(data)
    desc_num = data.describe().T
    desc_num.insert(0, 'variable', desc_num.index)
    desc_num.index = range(len(desc_num))
    cat_vars = set(data.columns) - set(desc_num["variable"]) - set(exclude_from_cat_desc)
    #cat_vars = data.columns - desc_num["variable"] - exclude_from_cat_desc
    desc_cat = None
    for var in cat_vars:
        c = data[var].value_counts()
        d = pd.DataFrame(dict(variable=var, level=c.index, count=c))[["variable", "level", "count"]]
        d.index = range(len(d))
        desc_cat = d if desc_cat is None else desc_cat.append(d)
    return basic, desc_num, desc_cat

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', help="Input csv file")
    parser.add_argument('-o', '--ouput', help="Input csv file")
    parser.add_argument('-e', '--exclude', help="variable to exclude (quoted, sep by space)")
    options = parser.parse_args()
    if not options.input or not options.ouput:
        parser.print_help()
        sys.exit(1)
    exclude = options.exclude
    if options.exclude is not None:
        exclude = options.exclude.split()
    else:
        exclude = []
    print(options.input, options.ouput, exclude)
    data = pd.read_csv(options.input)
    desc_num, desc_cat = describe_df(data, exclude=exclude)
    try:
        with pd.ExcelWriter(options.ouput) as writer:
            desc_num.to_excel(writer, sheet_name='Numercial')
            if desc_cat is not None:
                desc_cat.to_excel(writer, sheet_name='Categorial')
    except:
        out_num = options.ouput.replace(".xls", "_num.csv")
        desc_num.to_csv(out_num, index=False)
        if desc_cat is not None:
            out_cat = options.ouput.replace(".xls", "_cat.csv")
            desc_cat.to_csv(out_cat, index=False)



import pandas as pd
from sklearn.utils import shuffle
import random
from conpagnon.computing import compute_connectivity_matrices as ccm
from conpagnon.utils.array_operation import array_rebuilder
import copy
"""
This module contain useful operation on subjects connectivity matrices dictionnary.

Author: Dhaif BEKHA.
"""

# TODO: create a function for merging dictionnary key easily


def groupby_factor_connectivity_matrices(population_data_file, sheetname,
                                         subjects_connectivity_matrices_dictionnary,
                                         groupes, factors,
                                         drop_subjects_list=None,
                                         index_col=0):
    """Group by attribute the subjects connectivity matrices.
    # TODO: 18/09/2019: I added index_col to precise the index of the column
    # TODO: to be considered as the index of the whole dataframe.
    # TODO: Side Note: this function work with a time series dictionary too. !!
    # TODO: Refractoring of subjects_connectivity_matrices_dictionary to subjects_dictionary.
    """
    # Read the excel containing information of interest
    population_text_data = pd.read_excel(population_data_file, sheet_name=sheetname, index_col=index_col)

    # Drop subjects if needed
    if drop_subjects_list is not None:
        population_text_data = population_text_data.drop(drop_subjects_list)
    else:
        population_text_data = population_text_data

    all_subjects_list = []
    for groupe in groupes:
        for s in subjects_connectivity_matrices_dictionnary[groupe].keys():
            all_subjects_list.append(s)
    population_text_data = population_text_data.loc[population_text_data.index.intersection(all_subjects_list)]

    # Using pandas, group the dataframe by the factor list entered and store
    # it in a dictionary
    population_data_by_factor = population_text_data.groupby(factors).groups
    # Store all the keys, i.e all the possible factor pairs.
    factors_keys = list(population_data_by_factor.keys())

    # Create a subjects connectivity matrices dictionary, with factor keys as first level
    # keys, instead of groupes list
    group_by_factor_subjects_connectivity_matrices = dict.fromkeys(factors_keys)
    # Stack the connectivity matrices
    stacked_matrices = {s: subjects_connectivity_matrices_dictionnary[groupe][s] for groupe in groupes
                        for s in subjects_connectivity_matrices_dictionnary[groupe].keys()}
    # Fill the dictionary, with the corresponding subject level dictionary for each
    # factor key pair
    for factor in factors_keys:
        # subjects list ID for the current factor pair key
        subject_for_this_factor = list(population_data_by_factor[factor])
        group_by_factor_subjects_connectivity_matrices[factor] = dict.fromkeys(subject_for_this_factor)
        for s in subject_for_this_factor:
            group_by_factor_subjects_connectivity_matrices[factor][s] = stacked_matrices[s]

    return group_by_factor_subjects_connectivity_matrices, population_data_by_factor, factors_keys


def random_draw_of_connectivity_matrices(subjects_connectivity_dictionary, groupe, n_matrices,
                                         subjects_id_list=None, random_state=None, extract_kwargs=None):
    """Randomly pick N connectivity matrices from a subjects connectivity dictionary.

    Parameters
    ----------
    subjects_connectivity_dictionary: dict
        The subjects dictionary containing connectivity matrices
    groupe: str
        The group in which you want pick the matrices
    n_matrices: int
        The number of connectivity matrices you want
        to randomly choose
    subjects_id_list: list, optional
        The subjects identifiers list in which you
        want to choose matrices. If None, random matrices
        are picked in the entire group. Default is None.
    random_state: int, optional
        The seed of the pseudo random number generator to use when shuffling
        the data. If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the random number
        generator; If None, the random number generator is the RandomState
        instance used by `np.random`.
    extract_kwargs: dict, optional
        A dictionnary of argument passed to extract_sub_connectivity_matrices
        function. Default is None

    Returns
    -------
    output 1: dict
        The connectivity matrices dictionary, with subjects chosen
        randomly.
    output 2: list
        The  list of randomly chosen subjects identifier.

    """
    random_dictionary = {}
    # If a subjects list ids is entered
    if subjects_id_list is not None:
        subjects_id = list(set(subjects_id_list))
    else:
        # We take all the subjects ids
        subjects_id = list(set(subjects_connectivity_dictionary[groupe].keys()))

    # Shuffle the subjects id list, without impacting the original array
    shuffle_subjects_id = shuffle(subjects_id, random_state=random_state)
    # Randomly pick N subjects
    random_draw_subjects_id = random.sample(shuffle_subjects_id, n_matrices)
    # Extract the corresponding sub-dictionnary
    random_subjects_dictionary = {k: subjects_connectivity_dictionary[groupe][k]
                                  for k in random_draw_subjects_id}

    random_dictionary[groupe] = random_subjects_dictionary
    # if we want specific sub-matrices we extract them
    # extract sub connectivity matrices function
    if extract_kwargs is not None:
        random_dictionary = ccm.extract_sub_connectivity_matrices(
            subjects_connectivity_matrices=random_dictionary,
            kinds=extract_kwargs['kinds'],
            regions_index=extract_kwargs['regions_index'],
            vectorize=extract_kwargs['vectorize'],
            discard_diagonal=extract_kwargs['discard_diagonal'])

    return random_dictionary, random_draw_subjects_id


def merge_dictionary(dict_list, new_key=None):
    """Merge a list of dictionary

    Parameters
    ----------
    new_key: str, optional
        The key of the new merged dictionary. If None, the
        dictionaries in the list are simply merged together.
        Default is None
    dict_list: list
        A list of the dictionary to be merged

    Returns
    -------
    output: dict
        A dictionnary with one key, and merged dictionary
        as value.

    Notes
    -----
    Note that all the dictionnary you want to merge must have
    different keys.
    """
    merged_dictionary = dict()
    dictionary = {}
    for d in dict_list:
        merged_dictionary.update(d)
    if new_key is not None:
        dictionary[new_key] = merged_dictionary
    else:
        dictionary = merged_dictionary
    return dictionary


def stack_subjects_connectivity_matrices(subjects_connectivity_dictionary, groupes, kinds):
    """Re-arrange the subjects connectivity dictionary to return a stack version per group
    and kind.

    :param subjects_connectivity_dictionary:
    :param groupes:
    :param kinds:
    :return:
    """
    # Initialize dictionary
    stack_connectivity_dictionary = dict.fromkeys(groupes)

    for group in groupes:
        subjects_list = subjects_connectivity_dictionary[group].keys()
        stack_connectivity_dictionary[group] = dict.fromkeys(kinds)
        for kind in kinds:
            group_stacked_connectivity = [(s, subjects_connectivity_dictionary[group][s][kind])
                                          for s in subjects_list]
            group_stacked_mask = [(s, subjects_connectivity_dictionary[group][s]['masked_array'])
                                  for s in subjects_list]

            stack_connectivity_dictionary[group][kind] = {
                'connectivity matrices': group_stacked_connectivity,
                'masked_array': group_stacked_mask}

    return stack_connectivity_dictionary


def rebuild_subject_connectivity_matrices(subjects_connectivity_dictionary, groupes, kinds,
                                          diagonal_were_kept=False):
    """Given the subject connectivity dictionary, the matrix are rebuild from the vectorized
    one.

    Parameters
    ----------
    subjects_connectivity_dictionary: dict
        The subjects connectivity dictionary
    groupes: list
        The list of groups to rebuild the subjects matrices.
    kinds: list
        The list of kinds to rebuild.
    diagonal_were_kept: bool, optional
        If True, the reconstructed matrix, will have
        the diagonal store in the kind diagonal field of
        the dictionary, and the mask diagonal field for
        the mask.
        If False, the reconstructed matrix will have
        a zeros diagonal, and a True diagonal for the
        mask.

    Returns
    -------
    output 1: dict
        The reconstructed subjects connectivity
        matrices. All the matrices have now
        shape (number_of_regions, number_of_regions).

    Notes
    -----
    If in the input dictionary, the matrices and corresponding
    mask where vectorized with the diagonal kept, the argument
    `diagonal_is_there` must be set to False. A dimension
    error will be raises otherwise.

    """
    # Copy of the original dictionary to avoid side effect
    subjects_connectivity_dictionary_ = copy.deepcopy(subjects_connectivity_dictionary)

    for group in groupes:
        subjects_in_group = list(subjects_connectivity_dictionary_[group].keys())
        # First, we rebuild the vectorized mask
        for subject in subjects_in_group:
            vectorized_subject_mask = \
                subjects_connectivity_dictionary_[group][subject]['masked_array']
            # If the diagonal where kept in the vectorization process
            if diagonal_were_kept:
                # We simply rebuild the boolean mask
                rebuild_subject_mask = array_rebuilder(
                    vectorized_array=vectorized_subject_mask,
                    array_type='bool', diagonal=None)
            else:
                # We fetch the mask diagonal in the corresponding
                # subject dictionary
                subject_kind_mask_diagonal = \
                    subjects_connectivity_dictionary_[group][subject]['diagonal_mask']
                rebuild_subject_mask = array_rebuilder(
                    vectorized_array=vectorized_subject_mask,
                    array_type='bool', diagonal=subject_kind_mask_diagonal)
            # Now, iterate over the kind to rebuild the subject connectivity
            # matrices
            for kind in kinds:
                # Fetch the vectorized connectivity matrice of the current subject
                vectorized_subject_matrix = \
                    subjects_connectivity_dictionary_[group][subject][kind]
                # If the diagonal where kept in the vectorization process
                if diagonal_were_kept:
                    # Rebuild the connectivity matrices
                    rebuild_subject_matrix = array_rebuilder(
                        vectorized_array=vectorized_subject_matrix,
                        diagonal=None, array_type='numeric'
                    )
                else:
                    # Fetch the kind diagonal
                    subject_kind_diagonal = subjects_connectivity_dictionary_[group][subject][kind + '_diagonal']
                    # Rebuild the array
                    # override the present corresponding field
                    rebuild_subject_matrix = array_rebuilder(
                        vectorized_array=vectorized_subject_matrix,
                        diagonal=subject_kind_diagonal, array_type='numeric'
                    )

                subjects_connectivity_dictionary_[group][subject][kind] = rebuild_subject_matrix
                subjects_connectivity_dictionary_[group][subject]['masked_array'] = rebuild_subject_mask

    return subjects_connectivity_dictionary_



# -*- coding: utf-8 -*-
"""Display functions

Author: Dhaif BEKHA
"""

from nilearn.plotting import plot_connectome
import matplotlib.pyplot as plt
#import matplotlib
import seaborn as sns
import numpy as np
from matplotlib.backends import backend_pdf
from scipy import stats


def plot_ttest_results(t_test_dictionnary, groupes, contrast, node_coords, node_color,
                       output_pdf, node_size=50,
                       colormap='bwr', colorbar=False,
                       annotate=True,
                       display_mode='ortho'):
    """Plot on a glass brain the significant mean effect resulting of a
    two samples t-test and save resulting figures in a PDF files.
    
    Parameters
    ----------
    t_test_dictionnary : dict
        A dictionnary containing multiple keys :
            - 'tstatistic' : The raw statistic t-map for the choosen contrast
            2D numpy array of shape (number of regions, number of regions).
            
            - 'uncorrected pvalues' : The raw pvalues, 2D numpy array of shape
            (number of regions, number of regions).
            
            - 'corrected pvalues' : The corrected pvalues with the chosen methods.
            2D numpy array of shape (number of regions, number of regions).
            
            - 'significant edges' : The significant t-values after masking
            for the non significant pvalues at alpha level. 2D numpy array of shape 
            (number of regions, number of regions).
            
            - 'significant pvalues' : The significant pvalues at level alpha.
            2D numpy array of shape (number of regions, number of regions)
            
            - 'significant mean effect' : The differences of mean connectivity
            between the two groups according to the chosen contrast, and non-significant
            connexion are mask at alpha level. 2D numpy array of shape 
            (number of regions, number of regions).
    groupes : list
        The list of groupes involved in the computed
        contrast.
    contrast : list
        The contrast vector used in the two samples t-test.
    colormap: str, optional
        The colormap to use for the matrix heatmap.
        Default is 'bwr'.
    node_coords : list
        The list of centroids coordinates of each regions
        in the atlas used for regions signals extraction.
    node_color : numpy.ndarray shape(number of regions, 3)
        The color in the RGB normalized space, that is each
        color are represented by triplet of float ranging between
        0 and 1.
    output_pdf : str
        The full path including .pdf extension to the a pdf file
        for saving the glass brain projection plot.
    node_size : int, optional
        The node size. Default is 50.
    colorbar : bool, optional
        If True, the colorbar of the chosen colormap
        is displayed alongside the glass brain plot.
        Default is False.
    annotate : bool, optional
        If True, letters at the top of the glass brain
        indicates the left or right hemisphere location.
        Default is True.
    display_mode : str, optional
        The slice configuration you want to display
        Possible values are: 'ortho','x', 'y', 'z', 'xz' 
        'yx', 'yz', 'l', 'r', 'lr', 'lzr', 'lyr','lzry', 'lyrz'.
        Default if 'ortho'.
        
    See Also 
    --------
    parametric_tests.two_sample_t_test : 
        This function return a dictionnary resulting
        of the comparison of two groupe using a t-test. This
        the input of this display function.
        
    nilearn.plotting.plot_connectome :
        This function in the Nilearn packages is used here
        for the plot of the glass brain. I encourage
        the user to read the corresponding docstring in
        the Nilearn git repo.
        
    Notes
    -----
    The variables `groupes` and `contrast` are needed only
    the title. You shoul have already used this variables when
    you compute the two sample t-test between you're two groups.
    
    
    
    """

    # Modify the dictionnary of parameters for matplotlib
    import matplotlib.pylab as pylab
    params = {'legend.fontsize': 'x-large',
              'figure.figsize': (10, 4.5),
              'axes.labelsize': 'x-large',
              'axes.titlesize':'x-large',
              'xtick.labelsize':'x-large',
              'ytick.labelsize':'x-large'
              }
    pylab.rcParams.update(params)
    
    # For all kinds present in the t_test_dictionnary keys:
    with backend_pdf.PdfPages(output_pdf) as pdf:
        for kind in t_test_dictionnary.keys():
            # Write title for the current kind according contrast
            if contrast == [-1.0, 1.0]:
                title = 't-test for {} and contrast {} - {}'.format(kind, groupes[1], groupes[0])
            elif contrast == [1.0, -1.0]:
                title = 't-test for {} contrast {} - {}'.format(kind, groupes[0], groupes[1])
            else:
                raise ValueError('Unrecognized contrast vector !')

            # plot the connectome on a glass brain using Nilearn plot_connectome function
            fig = plt.figure()
            plot_connectome(adjacency_matrix=t_test_dictionnary[kind]['significant mean effect'],
                            node_coords=node_coords,
                            node_color=node_color, node_size=node_size,
                            edge_cmap=colormap, colorbar=colorbar,
                            annotate=annotate,
                            display_mode=display_mode, figure=fig, title=title)
            pdf.savefig()


def plot_matrix(matrix, labels_colors='auto', mpart='lower', k=0, colormap='RdBu_r',
                colorbar_params={'shrink': .5}, center=0,
                vmin=None, vmax=None, labels_size=8, horizontal_labels='auto',
                vertical_labels='auto',
                linewidths=.5, linecolor='white', title='Untitled',
                figure_size=(12, 9)):
    """Plot a entire matrix, or it's lower part with a chosen heatmap.
    
    Parameters
    ----------
    matrix : numpy.array shape(numbers of regions, numbers of regions)
        The numerical array you want to plot.
    labels_colors : str, or numpy.ndarray shape(number of regions, 3)
        The color in the RGB normalized space, that is each
        color are represented by triplet of float ranging between
        0 and 1. Default is 'auto', random colors are automatically
        generated.
    mpart : str
        The part of the array you want to plot, that is 
        the entire matrix or just it's lower part.
    k : int
        The position from the main diagonal you want 
        to cut. Default is 0.
    colormap : str
        The seaborn colormap you want to apply on the matrix
        when displaying it. Default is 'RdBu_r'.
    colorbar_params : dict, optional
        Additional parameters concerning the parameters
        of the colorbar : position, size...
        Keyword arguments for `fig.colorbar`.
    vmin, vmax : float, optional
        The extrema of the colormap. all display values
        are threshold according too those values. If None,
        these values are inferred from the matrix.
        Default is 'None'.
    labels_size : int, optional
        The font size of the labels.
        Default is 8.
    horizontal_labels : str or list; optional
        The labels list of each rows. If 'auto', labels
        are replaced by the row index.
    vertical_labels : str or list, optional
        The labels list of each columns. If 'auto', labels
        are replaced by the column index.
    linewidths : float, optional
        The width that divide each cell. Default
        is 0.5.
    linecolor : str, optional
        The color of the lines that divide each cell. Default
        is 'white'.
    title : str, optional
        The figure title. Default is 'untitled'.
    figure_size : tuple, optional
        Figure size in a tuple format (height, width).
        Default is (12, 9).
    center: float, optional
        The center of the colormap, for divergent data.
        
    See Also
    --------
    sns.heatmap :
        See the docstring of this function for a deeper
        insight of plotting function of a numerical array using
        seaborn.
        
    """
    
    # Compute matrices mask according to the chosen part of the matrix to plot.
    if mpart == 'lower':
        # Compute a boolean mask for the lower part of the array.
        matrice_mask = np.invert(np.ma.make_mask(np.tril(matrix, k=k)))
    elif mpart == 'all':
        # Compute a boolean  mask for all the array
        matrice_mask = np.invert(np.ma.make_mask(matrix))
    else:
        # Raise a value error, if the option are not among the authorized ones.
        raise ValueError('Unrecognized option for matrices mask.')
    # Generate random labels colors if labels_colors is set to auto:
    if labels_colors is 'auto':
        # Get numbers of labels
        n_labels = matrix.shape[0]
        labels_colors = np.random.rand(n_labels, 3)

    # plot matrices with the choosing cmap
    f, ax = plt.subplots(figsize=figure_size)
    for tick in ax.get_xticklabels():
        # Set label rotation to 90° for the x axis.
        tick.set_rotation(90)
    sns.heatmap(data=matrix, square=True, cmap=colormap, mask=matrice_mask, cbar_kws=colorbar_params,
                center=center,
                vmin=vmin, vmax=vmax, xticklabels=horizontal_labels,
                yticklabels=vertical_labels,
                linewidths=linewidths,
                linecolor=linecolor)
    # Set rotation angle to zero for the y axis.
    plt.yticks(rotation=0)
    for xtick, color in zip(ax.get_xticklabels(), labels_colors):
        xtick.set_color(color)
        xtick.set_fontsize(labels_size)
    for ytick, color in zip(ax.get_yticklabels(), labels_colors):
        ytick.set_color(color)
        ytick.set_fontsize(labels_size)
    plt.title(title)
    f.tight_layout()


def display_gaussian_connectivity_fit(vectorized_connectivity, estimate_mean, estimate_std,
                                      raw_data_colors='blue', line_width=2, alpha=0.5, normed=True,
                                      bins='auto', fitted_distribution_color='black', title=None,
                                      xtitle=None, ytitle=None, legend_fitted='Fitted Gaussian Distribution',
                                      legend_data=None, display_fit='yes', ms=5):
    """Display a vectorized connectivity matrices histogram, along with a Gaussian
    fit with an estimated mean, and standard deviation.

    Parameters
    ----------
    vectorized_connectivity: numpy.array, shape 0.5*n_features*(n_features + 1)
        A array of connectivity coefficient.
    estimate_mean: float
        The estimated mean of the connectivity coefficient distribution.
    estimate_std: float
        The estimated standard deviation of the connectivity coefficient distribution.
    raw_data_colors: string, optional
        The colors of the histogram for the connectivity coefficient distribution.
        Default is blue.
    line_width: int, optional
        The width of line for the fit of the data. Default is 2.
    alpha: float, optional
        The opacity coefficient for histogram, between 0 and 1. Default is 0.5.
    normed: bool, optional
        Normalized the histogram. This is mandatory for displaying
        the gaussian fit over the data, since it represent a probability density function.
    bins: string, optional
        How the bins edges of the histogram are computed, choices among different estimators
        {'fd', 'doane', 'scott', 'rice', 'sturges'}
        Default is 'auto', the maximum of the sturges and fd estimators are taken.
    fitted_distribution_color: string, optional
        The color of the curves representing the Gaussian fit. Default is black.
    title: string, optional
        The overall title of the figure. Default is None.
    xtitle: string, optional
        The legend of the x-axis. Default is None.
    ytitle: string, optional
        The legend of the y-axis. Default is None.
    legend_fitted: string, optional
        The legend for the Gaussian fit. Default is 'Fitted Gaussian Distribution'.
    legend_data: string, optional
        The legend for the histogram of the data. Default is None.
    display_fit: bool, optional
        If True, the gaussian fit of the data is displayed over the normalized histogram.
    ms: float, optional
        The size of the dot on the graph, default is 5.

    See Also
    --------
    matplotlib.pyplot.hist:
        This function, used here, compute and display the histogram of the data.

    """
    # Sort the vectorized vector of connectivity entered
    sorted_vectorized_connectivity = sorted(vectorized_connectivity)
    
    # Call matplotlib histogram function to draw the histogram of the data
    if display_fit is 'yes':
        distribution_density = stats.norm.pdf(sorted_vectorized_connectivity,
                                              estimate_mean, estimate_std)
        plt.plot(sorted_vectorized_connectivity, distribution_density, '-o',
                 lw=line_width, label=legend_fitted,
                 color=fitted_distribution_color, ms=ms)
        plt.legend(prop={'size': 7})
        plt.title(title)
        plt.xlabel(xtitle)
        plt.ylabel(ytitle)
    elif display_fit is 'no':
        count_array, bins_array, _ = plt.hist(sorted_vectorized_connectivity, label=legend_data,
                                              bins=bins, alpha=alpha, lw=line_width,
                                              color=raw_data_colors, normed=False)

        plt.legend(prop={'size': 7})
        plt.title(title)
        plt.xlabel(xtitle)
        plt.ylabel(ytitle)
    else:
        raise ValueError('Unrecognized display option ! '
                         'Choices are yes or no and you entered {}'.format(display_fit))


def t_and_p_values_barplot(t_values, p_values, alpha_level, xlabel_color, bar_labels,
                           t_xlabel, t_ylabel, p_xlabel, p_ylabel,
                           t_title, p_title, xlabel_size=10,
                           ylabel_size=10, size_label=8):
    """Plot a barplot representation of t and p values resulting
    of a two sample Student t-test in two separates figure.



    """
    # Conversion in numpy array format of t and p values list
    pvalues = np.array(p_values)
    tvalues = np.array(t_values)
    # Plot a seaborn bar plot for p-values

    # Put in grey the bar above the type I error threshold
    bar_color = [xlabel_color[i] if (pvalues[i] < alpha_level) else 'grey'
                 for i in range(len(bar_labels))]
    plt.figure(constrained_layout=True)
    ax = sns.barplot(x=bar_labels, y=pvalues, palette=bar_color)
    plt.xlabel(p_xlabel, size=xlabel_size)
    plt.ylabel(p_ylabel, size=ylabel_size)
    # Set font size for the x axis, and rotate the labels of 90° for visibility
    for xtick, color in zip(ax.get_xticklabels(), bar_color):
        xtick.set_color(color)
        xtick.set_fontsize(size_label)
        xtick.set_rotation(90)
    # Plot a asymptote for p-value above alpha threshold
    plt.hlines(y=alpha_level, xmin=ax.get_xlim()[0], xmax=ax.get_xlim()[1],
               label=str(alpha_level) + ' threshold', colors='red')
    plt.title(p_title)
    plt.legend()
    plt.show()
    # Plot the T statistic distribution
    plt.figure(constrained_layout=True)
    ax = sns.barplot(x=bar_labels, y=tvalues, palette=bar_color)
    plt.xlabel(t_xlabel, size=xlabel_size)
    plt.ylabel(t_ylabel, size=ylabel_size)
    for xtick, color in zip(ax.get_xticklabels(), bar_color):
        xtick.set_color(color)
        xtick.set_fontsize(size_label)
        xtick.set_rotation(90)
    plt.title(t_title)
   # plt.show()


def seaborn_scatterplot(x, y, data, figure_title, **kwargs):

    plt.figure()
    sns.lmplot(x=x, y=y, data=data, **kwargs)
    plt.title(figure_title)






"""
Modules to perform connectivity analysis at the network
level

Author: Dhaif BEKHA.

"""
from conpagnon.data_handling import dictionary_operations, data_management
from conpagnon.utils import folders_and_files_management
import os
from conpagnon.connectivity_statistics import parametric_tests
from conpagnon.pylearn_mulm import mulm
from patsy import dmatrix
import numpy as np
from statsmodels.sandbox.stats.multicomp import multipletests
from matplotlib.backends.backend_pdf import PdfPages
from conpagnon.plotting import display
import statsmodels.api as sm
import pandas as pd
import statsmodels.stats.multicomp as multi
from scipy import stats
from pathlib import Path


def regression_analysis_network_level(groups, kinds, networks_list, root_analysis_directory,
                                      network_model, variables_in_model, behavioral_dataframe,
                                      correction_method=['fdr_bh'], alpha=0.05,
                                      two_tailed=True, n_permutations=10000):
    """Regress a linear model at the network level
    """
    # The design matrix is the same for all model
    design_matrix = dmatrix('+'.join(variables_in_model), behavioral_dataframe, return_type='dataframe')
    for model in network_model:
        for kind in kinds:
            all_network_connectivity = []
            for network in networks_list:
                data_directory = os.path.join(root_analysis_directory,
                                              kind)
                # Concatenate all the intra-network dataframe
                network_model_dataframe = data_management.concatenate_dataframes([data_management.read_csv(
                    csv_file=os.path.join(data_directory, network, group + '_' + model + '_' +
                                          network + '_' + 'connectivity.csv')) for group in groups])
                # Shift index to be the subjects identifiers
                network_model_dataframe = data_management.shift_index_column(
                    panda_dataframe=network_model_dataframe,
                    columns_to_index=['subjects'])
                # Add variables in the model to complete the overall DataFrame
                network_model_dataframe = data_management.merge_by_index(
                    dataframe1=network_model_dataframe,
                    dataframe2=behavioral_dataframe[variables_in_model])
                # Build response variable vector, build matrix design
                network_response, network_design = parametric_tests.design_matrix_builder(
                    dataframe=network_model_dataframe,
                    formula=network_model_dataframe.columns[0] + '~' + '+'.join(variables_in_model)
                )

                # Fit the model
                network_model_fit = parametric_tests.ols_regression(y=network_response,
                                                                    X=network_design)

                # Creation of a directory for the current analysis for the current network
                regression_output_directory = folders_and_files_management.create_directory(
                    directory=os.path.join(root_analysis_directory,
                                           'regression_analysis', kind, network))

                # Write output regression results in csv files
                data_management.write_ols_results(ols_fit=network_model_fit, design_matrix=network_design,
                                                  response_variable=network_response,
                                                  output_dir=regression_output_directory,
                                                  model_name=model,
                                                  design_matrix_index_name='subjects')

                all_network_connectivity.append(network_response)

                # Take only the index subjects present in the analysis, because design matrix is for the whole
                # cohort !!!
                design_matrix = design_matrix.loc[network_design.index]

            # Multiple comparison correction

            # merge by index the dataframe from all the network for the current model
            all_network_response = data_management.merge_list_dataframes(all_network_connectivity)
            # Re-index the response variable dataframe to match the index of design matrix
            all_networks_connectivity = all_network_response.reindex(design_matrix.index)
            # Fit a linear model and correcting for maximum statistic
            mulm_fit = mulm.MUOLS(Y=np.array(all_networks_connectivity), X=np.array(design_matrix)).fit()
            # t-test for each variable in the model
            contrasts = np.identity(np.array(design_matrix).shape[1])
            raw_t, raw_p, df = mulm_fit.t_test(contrasts=contrasts, two_tailed=True, pval=True)
            for corr_method in correction_method:
                if corr_method == 'maxT':

                    _, p_values_maximum_T, _, null_distribution_max_T = \
                        mulm_fit.t_test_maxT(contrasts=contrasts, two_tailed=two_tailed,
                                             nperms=n_permutations)
                    corrected_p_values = p_values_maximum_T
                    # save the null distribution
                    folders_and_files_management.save_object(object_to_save=null_distribution_max_T,
                                                             saving_directory=os.path.join(root_analysis_directory,
                                                                                           'regression_analysis', kind),
                                                             filename=model+'_network_maximum_statistic_null.pkl')
                elif corr_method in ['fdr_bh', 'bonferroni']:
                    raw_p_shape = raw_p.shape
                    fdr_corrected_p_values = multipletests(pvals=raw_p.flatten(),
                                                           method=corr_method, alpha=alpha)[1].reshape(raw_p_shape)
                    corrected_p_values = fdr_corrected_p_values
                else:
                    # return raw p
                    corrected_p_values = raw_p

                # Append in each model for all networkCSV file,
                # the corrected p-values for the chosen
                # correction method
                for network in networks_list:
                    model_network_csv_file = os.path.join(root_analysis_directory,
                                                          'regression_analysis', kind,
                                                          network, model + '_parameters.csv')

                    # Read the csv file
                    model_network_parameters = data_management.read_csv(model_network_csv_file)
                    # Add a last column for adjusted p-values
                    model_network_parameters[corr_method + 'corrected_pvalues'] = \
                        corrected_p_values[:, networks_list.index(network)]
                    # Write it back to csf format
                    data_management.dataframe_to_csv(dataframe=model_network_parameters,
                                                     path=model_network_csv_file)


def regression_analysis_network_level_v2(groups, kinds, networks_list, root_analysis_directory,
                                         network_model, variables_in_model, score_of_interest,
                                         behavioral_dataframe,
                                         correction_method=['fdr_bh'],
                                         alpha=0.05):

    """Perform linear regression between a behavioral score and a functional connectivity "score".
    The connectivity score is usually a simple mean between the connectivity coefficient
    of a single network. ConPagnon has some function to compute those kind of score from
    a atlas, where at least a few network are identified from the user.

    Parameters
    ----------
    groups: list
        The list of string of the name of the
        group involved in the analysis. Usually,
        it's simply the entries of the subjects
        connectivity dictionary.
    kinds: list
        Repeat the statistical analysis for the
        all the connectivity metrics present
        in the list.
    networks_list: list
        Repeat the statistical analysis for
        all the network in the list. The network
        name should match the keys of the network
        dictionary containing the connectivity score.
    root_analysis_directory: str
        The full path to the text file containing
        all the text data to read and feed to the
        linear model.
    network_model: list
        A list containing the name of the model,
        matching the prefix of the corresponding
        text data to read.
    variables_in_model: list
        A list of the variables to put
        in the linear model. The variables
        name should match the columns name
        present in the text data.
    score_of_interest: str
        The behavioral score to analyze, it's
        simply the Y variable in the classical
        linear model: Y = X*Beta
    behavioral_dataframe: pandas.DataFrame
        The dataframe containing all the behavioral
        variables for each subjects.
    correction_method: list, optional
        A list containing all the desired correction method.
        Be careful, we do not stack all the models before
        the correction. By default, it's the Bonferonni
        method.
    alpha: float, optional
        The type I error rate, set to 0.05 by default.

    See Also
    --------
    compute_connectivity_matrices.intra_network_functional_connectivity :
        This function compute a intra-network connectivity score, by computing
        the mean of all connectivity coefficient inside a single network.
    """

    # The design matrix is the same for all model
    design_matrix = dmatrix('+'.join(variables_in_model), behavioral_dataframe, return_type='dataframe')
    for model in network_model:
        for kind in kinds:
            all_network_connectivity = []
            all_network_p_values = []
            for network in networks_list:
                data_directory = os.path.join(root_analysis_directory,
                                              kind)
                # Concatenate all the intra-network dataframe
                network_model_dataframe = data_management.concatenate_dataframes([data_management.read_csv(
                    csv_file=os.path.join(data_directory, network, group + '_' + model + '_' +
                                          network + '_' + 'connectivity.csv')) for group in groups])
                # Shift index to be the subjects identifiers
                network_model_dataframe = data_management.shift_index_column(
                    panda_dataframe=network_model_dataframe,
                    columns_to_index=['subjects'])

                connectivity_score_name = network_model_dataframe.columns[0]
                # fetch the score of interest in the behavioral dataframe
                score_dataframe = behavioral_dataframe[[score_of_interest]]
                # fetch the variables in the model, except connectivity score
                model_variables = behavioral_dataframe[variables_in_model]
                # Add variables in the model to complete the overall DataFrame
                network_model_dataframe = data_management.merge_list_dataframes(
                    [network_model_dataframe, score_dataframe,
                     model_variables])
                # Build the model formula: the variable to explain is the first column of the
                # dataframe, and we add to the left all variable in the model
                model_formulation = score_of_interest + '~' + '+'.join(variables_in_model + [connectivity_score_name])
                # Build response variable vector, build matrix design
                network_response, network_design = parametric_tests.design_matrix_builder(
                    dataframe=network_model_dataframe,
                    formula=model_formulation)

                # Fit the model
                network_model_fit = parametric_tests.ols_regression(y=network_response,
                                                                    X=network_design)

                # Creation of a directory for the current analysis for the current network
                regression_output_directory = folders_and_files_management.create_directory(
                    directory=os.path.join(root_analysis_directory,
                                           'regression_analysis', kind, network))

                # Write output regression results in csv files
                data_management.write_ols_results(ols_fit=network_model_fit, design_matrix=network_design,
                                                  response_variable=network_response,
                                                  output_dir=regression_output_directory,
                                                  model_name=model,
                                                  design_matrix_index_name='subjects')

                all_network_connectivity.append(network_response)
                all_network_p_values.append(network_model_fit.pvalues)

                # Take only the index subjects present in the analysis, because design matrix is for the whole
                # cohort !!!
                design_matrix = design_matrix.loc[network_design.index]

            # Multiple comparison correction
            # merge by index the dataframe from all the network for the current model
            all_network_response = data_management.merge_list_dataframes(all_network_connectivity)
            # Re-index the response variable dataframe to match the index of design matrix
            all_networks_connectivity = all_network_response.reindex(design_matrix.index)
            # t-test for each variable in the model
            contrasts = np.identity(np.array(design_matrix).shape[1])
            raw_p = np.array(all_network_p_values).T
            for corr_method in correction_method:
                if corr_method in ['fdr_bh', 'bonferroni']:
                    raw_p_shape = raw_p.shape
                    fdr_corrected_p_values = multipletests(pvals=raw_p.flatten(),
                                                           method=corr_method, alpha=alpha)[1].reshape(raw_p_shape)
                    corrected_p_values = fdr_corrected_p_values
                else:
                    # return raw p
                    corrected_p_values = raw_p

                # Append in each model for all networkCSV file,
                # the corrected p-values for the chosen
                # correction method
                for network in networks_list:
                    model_network_csv_file = os.path.join(root_analysis_directory,
                                                          'regression_analysis', kind,
                                                          network, model + '_parameters.csv')

                    # Read the csv file
                    model_network_parameters = data_management.read_csv(model_network_csv_file)
                    # Add a last column for adjusted p-values
                    model_network_parameters[corr_method + 'corrected_pvalues'] = \
                        corrected_p_values[:, networks_list.index(network)]
                    # Write it back to csf format
                    data_management.dataframe_to_csv(dataframe=model_network_parameters,
                                                     path=model_network_csv_file)


def regression_analysis_whole_brain(groups, kinds, root_analysis_directory,
                                    whole_brain_model, variables_in_model,
                                    behavioral_dataframe,
                                    correction_method=['FDR'], alpha=0.05,
                                    n_permutations=10000,
                                    two_tailed=True):
    """Regression analysis on composite connectivity score over the whole brain.

    """

    # The design matrix is the same for all model
    design_matrix = dmatrix('+'.join(variables_in_model), behavioral_dataframe, return_type='dataframe')
    # For each model: read the csv for each group, concatenate resultings dataframe, and append
    # (merging by index) all variable of interest in the model.
    for kind in kinds:
        # directory where the data are
        data_directory = os.path.join(root_analysis_directory,
                                      kind)
        all_model_response = []
        for model in whole_brain_model:
            # List of the corresponding dataframe
            model_dataframe = data_management.concatenate_dataframes([data_management.read_csv(
                csv_file=os.path.join(data_directory, group + '_' + kind + '_' + model + '.csv'))
                for group in groups])
            # Shift index to be the subjects identifiers
            model_dataframe = data_management.shift_index_column(panda_dataframe=model_dataframe,
                                                                 columns_to_index=['subjects'])
            # Add variables in the model to complete the overall DataFrame
            model_dataframe = data_management.merge_by_index(dataframe1=model_dataframe,
                                                             dataframe2=behavioral_dataframe[variables_in_model])
            # Build the model formula: the variable to explain is the first column of the
            # dataframe, and we add to the left all variable in the model
            model_formulation = model_dataframe.columns[0] + '~' + '+'.join(variables_in_model)
            # Build response, and design matrix from the model model formulation
            model_response, model_design = \
                parametric_tests.design_matrix_builder(dataframe=model_dataframe,
                                                       formula=model_formulation,
                                                       return_type='dataframe')
            # regression with a simple OLS model
            model_fit = parametric_tests.ols_regression(y=model_response, X=model_design)

            # Creation of a directory for the current analysis
            regression_output_directory = folders_and_files_management.create_directory(
                directory=os.path.join(root_analysis_directory, 'regression_analysis', kind))

            # Write output regression results in csv files
            data_management.write_ols_results(ols_fit=model_fit, design_matrix=model_design,
                                              response_variable=model_response,
                                              output_dir=regression_output_directory,
                                              model_name=model,
                                              design_matrix_index_name='subjects')
            # Appending current model response
            all_model_response.append(model_response)

            # Take only the index subjects present in the analysis, because design matrix is for the whole
            # cohort !!!
            design_matrix = design_matrix.loc[model_design.index]

        # merge by index the dataframe
        df_tmp = data_management.merge_list_dataframes(all_model_response)
        # Re-index the response variable dataframe to match the index of design matrix
        model_whole_brain_connectivity = df_tmp.reindex(design_matrix.index)

        # Fit a linear model and correcting for maximum statistic
        mulm_fit = mulm.MUOLS(Y=np.array(model_whole_brain_connectivity),
                              X=np.array(design_matrix)).fit()
        contrasts = np.identity(np.array(design_matrix).shape[1])
        raw_t, raw_p, df = mulm_fit.t_test(contrasts=contrasts, two_tailed=True, pval=True)
        for corr_method in correction_method:

            if corr_method == 'maxT':

                _, p_values_maximum_T, _, null_distribution_max_T = \
                    mulm_fit.t_test_maxT(contrasts=contrasts, two_tailed=two_tailed,
                                         nperms=n_permutations)
                corrected_p_values = p_values_maximum_T
                # save the null distribution
                folders_and_files_management.save_object(
                    object_to_save=null_distribution_max_T,
                    saving_directory=os.path.join(root_analysis_directory,
                                                  'regression_analysis', kind),
                                                  filename='whole_brain_connectivity_maximum_statistic_null.pkl')
            elif corr_method in ['fdr_bh', 'bonferroni']:
                raw_p_shape = raw_p.shape
                fdr_corrected_p_values = multipletests(pvals=raw_p.flatten(),
                                                       method=corr_method, alpha=alpha)[1].reshape(raw_p_shape)
                corrected_p_values = fdr_corrected_p_values
            else:
                corrected_p_values = raw_p

            # Append in each model CSV file, the corrected p-values for maximum statistic
            for model in whole_brain_model:
                model_csv_file = os.path.join(root_analysis_directory, 'regression_analysis', kind,
                                              model + '_parameters.csv')
                # Read the csv file
                model_parameters = data_management.read_csv(model_csv_file)
                # Add a last column for adjusted p-values
                model_parameters[corr_method + 'corrected_pvalues'] = \
                    corrected_p_values[:, whole_brain_model.index(model)]
                # Write it back to csf format
                data_management.dataframe_to_csv(dataframe=model_parameters, path=model_csv_file)


def regression_analysis_whole_brain_v2(groups, kinds, root_analysis_directory,
                                       whole_brain_model, variables_in_model, score_of_interest,
                                       behavioral_dataframe,
                                       correction_method=['fdr_bh'],
                                       alpha=0.05):
    """Compute a linear model with a continuous score test as the response variable.

    :param groups:
    :param kinds:
    :param root_analysis_directory:
    :param whole_brain_model:
    :param variables_in_model:
    :param score_of_interest:
    :param behavioral_dataframe:
    :param correction_method:
    :param alpha:
    :return:
    """
    for kind in kinds:
        # directory where the data are
        data_directory = os.path.join(root_analysis_directory,
                                      kind)
        all_model_response = []
        all_model_p_values = []
        for model in whole_brain_model:
            # List of the corresponding dataframe
            model_dataframe = data_management.concatenate_dataframes([data_management.read_csv(
                csv_file=os.path.join(data_directory, group + '_' + kind + '_' + model + '.csv'))
                for group in groups])

            # Shift index to be the subjects identifiers
            model_dataframe = data_management.shift_index_column(panda_dataframe=model_dataframe,
                                                                 columns_to_index=['subjects'])
            # Fetch the name of the columns containing the
            # connectivity score. There is only one element
            # since we shifted the index to subject list
            # previously.
            connectivity_score_name = model_dataframe.columns[0]
            # fetch the score of interest in the behavioral dataframe
            score_dataframe = behavioral_dataframe[[score_of_interest]]
            # fecth the variables in the model, except connectivity score
            model_variables = behavioral_dataframe[variables_in_model]
            # Add variables in the model to complete the overall DataFrame
            model_dataframe = data_management.merge_list_dataframes([model_dataframe, score_dataframe, model_variables])
            # Build the model formula: the variable to explain is the first column of the
            # dataframe, and we add to the left all variable in the model
            model_formulation = score_of_interest + '~' + '+'.join(variables_in_model + [connectivity_score_name])
            # Build response, and design matrix from the model model formulation
            model_response, model_design = \
                parametric_tests.design_matrix_builder(dataframe=model_dataframe,
                                                       formula=model_formulation,
                                                       return_type='dataframe')
            # regression with a simple OLS model
            model_fit = parametric_tests.ols_regression(y=model_response, X=model_design)

            # Creation of a directory for the current analysis
            regression_output_directory = folders_and_files_management.create_directory(
                directory=os.path.join(root_analysis_directory, 'regression_analysis', kind))

            # Write output regression results in csv files
            data_management.write_ols_results(ols_fit=model_fit, design_matrix=model_design,
                                              response_variable=model_response,
                                              output_dir=regression_output_directory,
                                              model_name=model,
                                              design_matrix_index_name='subjects')
            # Appending current model response
            all_model_response.append(model_response)

            # Take only the index subjects present in the analysis, because design matrix is for the whole
            # cohort !!!
            design_matrix = model_design.loc[model_design.index]

            # append p values
            all_model_p_values.append(np.array(model_fit.pvalues))

        raw_p = np.array(all_model_p_values).T
        for corr_method in correction_method:
            if corr_method in ['fdr_bh', 'bonferroni']:
                raw_p_shape = raw_p.shape
                fdr_corrected_p_values = multipletests(pvals=raw_p.flatten(),
                                                       method=corr_method, alpha=alpha)[1].reshape(raw_p_shape)
                corrected_p_values = fdr_corrected_p_values
            else:
                corrected_p_values = raw_p

            # Append in each model CSV file, the corrected p-values for maximum statistic
            for model in whole_brain_model:
                model_csv_file = os.path.join(root_analysis_directory, 'regression_analysis', kind,
                                              model + '_parameters.csv')
                # Read the csv file
                model_parameters = data_management.read_csv(model_csv_file)
                # Add a last column for adjusted p-values
                model_parameters[corr_method + 'corrected_pvalues'] = \
                    corrected_p_values[:, whole_brain_model.index(model)]
                # Write it back to csf format
                data_management.dataframe_to_csv(dataframe=model_parameters, path=model_csv_file)


def regression_analysis_internetwork_level(internetwork_subjects_connectivity_dictionary, groups_in_model,
                                           behavioral_data_path,
                                           sheet_name, subjects_to_drop, model_formula, kinds_to_model,
                                           root_analysis_directory,
                                           inter_network_model, network_labels_list, network_labels_colors,
                                           pvals_correction_method=['fdr_bh'], vectorize=True,
                                           discard_diagonal=False, nperms_maxT = 10000, contrasts = 'Id',
                                           compute_pvalues = 'True', pvalues_tail = 'True', NA_action='drop',
                                           alpha=0.05
                                           ):
    
    # Merge the dictionary into one, to give it as data argument to
    # linear regression function
    group_dictionary_list = [internetwork_subjects_connectivity_dictionary[group] for group in groups_in_model]
    connectivity_data = dictionary_operations.merge_dictionary(group_dictionary_list,
                                                               'all_subjects')['all_subjects']
    
    # We iterate for each kind
    for kind in kinds_to_model:
        # Create and save the raw data dictionary 
        raw_connectivity_data_output = data_management.create_directory(
                directory=os.path.join(root_analysis_directory, kind, inter_network_model))
        
        folders_and_files_management.save_object(object_to_save=connectivity_data,
                                                 saving_directory=raw_connectivity_data_output,
                                                 filename=inter_network_model + 'data.pkl'
                                                 )
        # Create the output directory for the current kind
        inter_network_analysis_output = data_management.create_directory(
                directory=os.path.join(root_analysis_directory, 'regression_analysis',
                                       kind, 
                                       inter_network_model))
        
        regression_results, X_df, y, y_prediction, regression_subjects_list = parametric_tests.linear_regression(
            connectivity_data=connectivity_data,
            data=behavioral_data_path,
            sheetname=sheet_name,
            subjects_to_drop=subjects_to_drop,
            pvals_correction_method=pvals_correction_method,
            save_regression_directory=inter_network_analysis_output,
            kind=kind,
            NA_action=NA_action,
            formula=model_formula,
            vectorize=vectorize,
            discard_diagonal=discard_diagonal,
            nperms_maxT=nperms_maxT,
            contrasts=contrasts,
            compute_pvalues=compute_pvalues,
            pvalues_tail=pvalues_tail)
        # Save the design matrix 
        data_management.dataframe_to_csv(dataframe=X_df,
                                         path=os.path.join(inter_network_analysis_output, 'design_matrix.csv'),
                                         index=True)
        
        # Create the figure for each correction method: matrix plot
        # of significant p-values, and t-values
        
        # For each variable plot and save a matrix, with T statistic masked
        # for significant values, and significant p values for each variables
        corr_method_output_pdf = os.path.join(inter_network_analysis_output,
                                              kind + '_' + inter_network_model + '.pdf')
        
        with PdfPages(corr_method_output_pdf) as pdf:
            
            for corr_method in pvals_correction_method:
                # Fetch the results of the model 
                corr_method_results = regression_results[corr_method]['results']
                # iterate over all the keys of dictionary results, i.e
                # all variable
                for variable in corr_method_results.keys():  
                    # Significant t statistic
                    display.plot_matrix(matrix=corr_method_results[variable]['significant tvalues'],
                                        labels_colors=network_labels_colors, mpart='lower', k=-1,
                                        colormap='RdBu_r',
                                        colorbar_params={'shrink': .5}, center=0,
                                        vmin=None, vmax=None, labels_size=8,
                                        horizontal_labels=network_labels_list,
                                        vertical_labels=network_labels_list,
                                        linewidths=.5, linecolor='black',
                                        title=corr_method + '_' + variable + '_effect',
                                        figure_size=(12, 9))
                    pdf.savefig()
                    # Significant p values
                    display.plot_matrix(matrix=corr_method_results[variable]['corrected pvalues'],
                                        labels_colors=network_labels_colors, mpart='lower', k=-1,
                                        colormap='hot', colorbar_params={'shrink': .5}, center=0,
                                        vmin=0, vmax=alpha, labels_size=8,
                                        horizontal_labels=network_labels_list,
                                        vertical_labels=network_labels_list,
                                        linewidths=.5, linecolor='black',
                                        title=corr_method + '_' + variable + '_p values',
                                        figure_size=(12, 9))
                    pdf.savefig()


def write_raw_data(output_csv_directory, kinds, groups, models, variables_in_model,
                   behavioral_dataframe):
    """Write the response variable along with the co-variables in a unique
    dataframe inside a csv files.

    Parameters
    ----------
    output_csv_directory: str
        The directory path containing the data for each connectivity
        metric. It should contain one folder for each metric in you're analysis.
    kinds: list
        The list of connectivity metric needed in the statistical analysis.
    groups: list
        The list of subjects group names needed in the statistical analysis.
    models: list
        The list of models needed in the analysis. The model name should be
        contain in the group csv data.
    variables_in_model: list
        The list of co-variables to add to the response variable.
    behavioral_dataframe: pandas.DataFrame
        A pandas dataframe containing the co-variables to add. Note
        that the dataframe must have a columns called 'subjects' with the
        identifier of each subjects. The response variable dataframe and
        the behavioral dataframe will be merge considering the subject
        columns as index.

    """
    # Read and concatenate list of the data for each group
    for model in models:
        for kind in kinds:
            # Read and concatenate the data for each group to build the response variable
            response_variable_data = data_management.concatenate_dataframes([data_management.read_csv(
                csv_file=os.path.join(output_csv_directory, kind, group + '_' + kind + '_' + model +
                                      '.csv')) for group in groups])
            # Make the 'subjects' column the index of the response variable dataframe
            response_variable_data = data_management.shift_index_column(
                panda_dataframe=response_variable_data,
                columns_to_index=['subjects'])
            # Merge the response variable dataframe and the behavioral
            # dataframe to build the overall dataframe
            model_dataframe = data_management.merge_by_index(
                dataframe1=response_variable_data,
                dataframe2=behavioral_dataframe[variables_in_model])

            model_dataframe.to_csv(os.path.join(output_csv_directory, kind, model + '_raw_data.csv'))


def write_raw_data_network(output_csv_directory, kinds, groups, models, variables_in_model,
                           behavioral_dataframe, network_list):
    """Write the response variable along with the co-variables in a unique
    dataframe inside a csv files, at the network level.

    Parameters
    ----------
    output_csv_directory: str
        The directory path containing the data for each connectivity
        metric. It should contain one folder for each metric in you're analysis.
    kinds: list
        The list of connectivity metric needed in the statistical analysis.
    groups: list
        The list of subjects group names needed in the statistical analysis.
    models: list
        The list of models needed in the analysis. The model name should be
        contain in the group csv data.
    variables_in_model: list
        The list of co-variables to add to the response variable.
    behavioral_dataframe: pandas.DataFrame
        A pandas dataframe containing the co-variables to add. Note
        that the dataframe must have a columns called 'subjects' with the
        identifier of each subjects. The response variable dataframe and
        the behavioral dataframe will be merge considering the subject
        columns as index.
    network_list: list
        The list of network names. Each network  in each connectivity metric
        should have a directory containing the data for each group.
    """
    for network in network_list:
        # Read and concatenate list of the data for each group
        for model in models:
            for kind in kinds:
                # Read and concatenate the data for each group to build the response variable
                response_variable_data = data_management.concatenate_dataframes([data_management.read_csv(
                    csv_file=os.path.join(output_csv_directory, kind, network, group + '_' + model + '_' + network +
                                          '_connectivity.csv')) for group in groups])
                # Make the 'subjects' column the index of the response variable dataframe
                response_variable_data = data_management.shift_index_column(
                    panda_dataframe=response_variable_data,
                    columns_to_index=['subjects'])
                # Merge the response variable dataframe and the behavioral
                # dataframe to build the overall dataframe
                model_dataframe = data_management.merge_by_index(
                    dataframe1=response_variable_data,
                    dataframe2=behavioral_dataframe[variables_in_model])

                model_dataframe.to_csv(os.path.join(output_csv_directory, kind, network, model + '_raw_data.csv'))


def one_way_anova(models, groups, behavioral_dataframe, kinds,
                  correction_method, root_analysis_directory,
                  variables_in_model,
                  alpha=.05):
    """Perform one way ANOVA analysis followed by a post hoc analysis with t test,
    and multiple comparison correction.

    Parameters
    ----------
    models: list
        List of models. The list model is defined as the string in the filename
        containing the raw data for each groups.
    groups: list
        The list of group in the study.
    behavioral_dataframe: pandas.DataFrame
        The dataframe containing the variable to study in the ANOVA analysis.
        The dataframe must contain a column named "subjects" containing the
        identifier for each subjects.
    kinds: list
        The list of the different connectivity metrics you want to
        perform the analysis.
    correction_method: list
        The list of multiple comparison correction method, as available
        in the statsmodels library.
    root_analysis_directory: string
        The full path to the directory containing the raw data.
    variables_in_model: list
        A list containing the categorical variable to study.
    alpha: float, optional
        The type I error threshold. The default is 0.05.

    Notes
    -----
    If models in a list containing more than one elements,
    the p values are jointly corrected for all models.

    """
    # For each model: read the csv for each group, concatenate resulting dataframe, and append
    # (merging by index) all variable of interest in the model.
    for kind in kinds:
        # directory where the data are
        data_directory = os.path.join(root_analysis_directory,
                                      kind)
        # Store the p value of all the tested model for multiple
        # comparison correction
        all_model_p_values_f_test = []
        # Store the p values resulting in all possible contrast
        # t test for multiple comparison correction
        all_model_post_hoc_p_values = []

        # Creation of a directory for the current analysis, ANOVA results
        # will be stored in a sub-directory called ANOVA
        regression_output_directory = folders_and_files_management.create_directory(
            directory=os.path.join(root_analysis_directory, 'regression_analysis/ANOVA', kind))
        for model in models:
            # List of the corresponding dataframe
            model_dataframe = data_management.concatenate_dataframes([data_management.read_csv(
                csv_file=os.path.join(data_directory, group + '_' + kind + '_' + model + '.csv'))
                for group in groups])
            # Shift index to be the subjects identifiers
            model_dataframe = data_management.shift_index_column(panda_dataframe=model_dataframe,
                                                                 columns_to_index=['subjects'])
            # Add variables in the model to complete the overall DataFrame
            model_dataframe = data_management.merge_by_index(dataframe1=model_dataframe,
                                                             dataframe2=behavioral_dataframe[variables_in_model])
            # Build the model formula: the variable to explain is the first column of the
            # dataframe, and we add to the left all variable in the model
            model_formulation = model_dataframe.columns[0] + '~' + '+'.join(variables_in_model)
            # regression with a simple OLS model
            model_fit = parametric_tests.ols_regression_formula(formula=model_formulation,
                                                                data=model_dataframe)

            # Perform one-way ANOVA with statsmodel
            model_anova = sm.stats.anova_lm(model_fit, typ=2)
            # Compute t-test between all pairs of groups in the
            # analysis with tukey HSD analysis
            model_tukey_analysis = multi.MultiComparison(
                model_dataframe[model_dataframe.columns[0]],
                model_dataframe[variables_in_model[0]])
            model_tukey_analysis_results = model_tukey_analysis.tukeyhsd()
            # tukey t-score are t score scaled by sqrt(2)
            model_t_scores = \
                model_tukey_analysis_results.meandiffs / model_tukey_analysis_results.std_pairs / np.sqrt(2)
            # Compute two-sided uncorrected p values
            model_raw_p_values = stats.t.sf(np.abs(model_t_scores), 50)*2

            # Stack the p values for the F test of the current model
            all_model_p_values_f_test.append(model_anova['PR(>F)'].loc[variables_in_model[0]])
            # Stack the p values for all contrast for the current model
            all_model_post_hoc_p_values.append(model_raw_p_values)

            # In the report the contrast is group 2 - group 1, build a dataframe to store
            # the results
            post_hoc_dataframe = pd.DataFrame(data=model_tukey_analysis_results._results_table.data[1:],
                                              columns=model_tukey_analysis_results._results_table.data[0])
            # Add t scores for each contrast and raw p values
            post_hoc_dataframe = post_hoc_dataframe.assign(t=model_t_scores, p_values=model_raw_p_values)
            # Save: the results containing the F value and p value for the model
            model_anova.to_csv(os.path.join(regression_output_directory,
                                            model + '_f_test_parameters.csv'),
                               index=False)
            # Save the parameters results of the post hoc analysis
            data_management.dataframe_to_csv(dataframe=post_hoc_dataframe,
                                             path=os.path.join(regression_output_directory,
                                                               model + '_post_hoc_parameters.csv'),
                                             index=False)

        # Correct the p values for all model, and post-hoc t test

        # The user can choose different correction method
        p_values_correction_post_hoc_all_method = np.zeros((len(correction_method), len(models),
                                                            np.array(all_model_post_hoc_p_values).shape[1]))
        p_values_correction_f_test_all_method = np.zeros((len(correction_method), len(models)))
        for correction in correction_method:

            # Correction with the chosen method for the p value of all between group t test
            all_model_p_values_post_hoc_array = np.array(all_model_post_hoc_p_values)
            post_hoc_t_test_corrected_p_values = multipletests(
                pvals=all_model_p_values_post_hoc_array.flatten(),
                method=correction,
                alpha=alpha)[1].reshape(all_model_p_values_post_hoc_array.shape)

            p_values_correction_post_hoc_all_method[correction_method.index(correction), ...] = \
                post_hoc_t_test_corrected_p_values

            # Correction with the chosen method for the p value of the whole model
            all_model_p_values_f_test_array = np.array(all_model_p_values_f_test)
            f_test_corrected_p_values = multipletests(
                pvals=all_model_p_values_f_test_array.flatten(),
                method=correction,
                alpha=alpha)[1].reshape(all_model_p_values_f_test_array.shape)

            p_values_correction_f_test_all_method[correction_method.index(correction), ...] =\
                f_test_corrected_p_values

        # Finally we can now append the corrected p value for all the chosen method in
        # for all model in the csv file
        for model in models:
            # Read the parameters file of the model for the f test first
            f_test_model_parameters = data_management.read_csv(os.path.join(
                regression_output_directory, model + '_f_test_parameters.csv'))
            # Add the corrected p values for all the correction method
            post_hoc_model_parameters = data_management.read_csv(os.path.join(
                regression_output_directory, model + '_post_hoc_parameters.csv'))

            # Append the corrected p values
            for correction in correction_method:
                corrected_p_value_column_name = correction + '_p_value'
                post_hoc_model_parameters[corrected_p_value_column_name] = \
                    p_values_correction_post_hoc_all_method[correction_method.index(correction),
                                                            models.index(model), ...]

                f_test_model_parameters[corrected_p_value_column_name] = \
                    p_values_correction_f_test_all_method[correction_method.index(correction),
                                                          models.index(model)]

                post_hoc_model_parameters.to_csv(os.path.join(regression_output_directory,
                                                              model + '_post_hoc_parameters.csv'),
                                                 index=False)

                f_test_model_parameters.to_csv(os.path.join(regression_output_directory,
                                                            model + '_f_test_parameters.csv'),
                                               index=False)


def one_way_anova_network(root_analysis_directory, kinds, groups,
                          networks_list, models, behavioral_dataframe,
                          variables_in_model, correction_method, alpha=0.05):
    """Perform a one way ANOVA at the network level.

    Parameters
    ----------
    models: list
        List of models. The list model is defined as the string in the filename
        containing the raw data for each groups.
    groups: list
        The list of group in the study.
    networks_list: list
        The list of network to include in the analysis
    behavioral_dataframe: pandas.DataFrame
        The dataframe containing the variable to study in the ANOVA analysis.
        The dataframe must contain a column named "subjects" containing the
        identifier for each subjects.
    kinds: list
        The list of the different connectivity metrics you want to
        perform the analysis.
    correction_method: list
        The list of multiple comparison correction method, as available
        in the statsmodels library.
    root_analysis_directory: string
        The full path to the directory containing the raw data.
    variables_in_model: list
        A list containing the categorical variable to study.
    alpha: float, optional
        The type I error threshold. The default is 0.05.

    Notes
    -----
    If models  and network_list is a list
    containing more than one elements, the p values
    are jointly corrected for the number of models * number of networks.

    """

    for kind in kinds:
        # List to stack the p values from the F test of the model
        all_network_f_test_p_values = []
        # List to stack the p values from all possible contrast from
        # the post hoc t test
        all_network_post_hoc_test_p_values = []
        for network in networks_list:
            for model in models:
                # Creation of a directory for the current analysis, ANOVA results
                # will be stored for each network
                regression_output_directory = folders_and_files_management.create_directory(
                    directory=os.path.join(root_analysis_directory, 'regression_analysis/ANOVA', kind,
                                           network))

                network_data_directory = os.path.join(root_analysis_directory, kind, network)
                # Concatenate all the intra-network dataframe
                network_model_dataframe = data_management.concatenate_dataframes([data_management.read_csv(
                    csv_file=os.path.join(network_data_directory, group + '_' + model + '_' +
                                          network + '_' + 'connectivity.csv')) for group in groups])
                # Shift index to be the subjects identifiers
                network_model_dataframe = data_management.shift_index_column(
                    panda_dataframe=network_model_dataframe,
                    columns_to_index=['subjects'])
                # Add variables in the model to complete the overall DataFrame
                network_model_dataframe = data_management.merge_by_index(
                    dataframe1=network_model_dataframe,
                    dataframe2=behavioral_dataframe[variables_in_model])
                # Build response variable vector, build matrix design
                network_response, network_design = parametric_tests.design_matrix_builder(
                    dataframe=network_model_dataframe,
                    formula=network_model_dataframe.columns[0] + '~' + '+'.join(variables_in_model)
                )
                # Build the model formula: the variable to explain is the first column of the
                # dataframe, and we add to the left all variable in the model
                model_formulation = network_model_dataframe.columns[0] + '~' + '+'.join(variables_in_model)
                # regression with a simple OLS model
                network_model_fit = parametric_tests.ols_regression_formula(formula=model_formulation,
                                                                            data=network_model_dataframe)

                # Perform one-way ANOVA with statsmodel
                network_model_anova = sm.stats.anova_lm(network_model_fit, typ=2)
                # Compute t-test between all pairs of groups in the
                # analysis with tukey HSD analysis for the current network
                network_model_tukey_analysis = multi.MultiComparison(
                    network_model_dataframe[network_model_dataframe.columns[0]],
                    network_model_dataframe[variables_in_model[0]])
                network_model_tukey_analysis_results = network_model_tukey_analysis.tukeyhsd()
                # tukey t-score are t score scaled by sqrt(2)
                network_model_t_scores = \
                    network_model_tukey_analysis_results.meandiffs / network_model_tukey_analysis_results.std_pairs / np.sqrt(2)
                # Compute two-sided uncorrected p values
                network_model_raw_p_values = stats.t.sf(np.abs(network_model_t_scores), 50) * 2

                # Stack the p values for the F test of the current model
                all_network_f_test_p_values.append(network_model_anova['PR(>F)'].loc[variables_in_model[0]])
                # Stack the p values for all contrast for the current model
                all_network_post_hoc_test_p_values.append(network_model_raw_p_values)

                # In the report the contrast is group 2 - group 1, build a dataframe to store
                # the results
                post_hoc_dataframe = pd.DataFrame(data=network_model_tukey_analysis_results._results_table.data[1:],
                                                  columns=network_model_tukey_analysis_results._results_table.data[0])
                # Add t scores for each contrast and raw p values
                post_hoc_dataframe = post_hoc_dataframe.assign(t=network_model_t_scores,
                                                               p_values=network_model_raw_p_values)
                # Save: the results containing the F value and p value for the model
                network_model_anova.to_csv(os.path.join(regression_output_directory,
                                                        model + '_f_test_parameters.csv'),
                                           index=False)
                # Save the parameters results of the post hoc analysis
                data_management.dataframe_to_csv(dataframe=post_hoc_dataframe,
                                                 path=os.path.join(regression_output_directory,
                                                                   model + '_post_hoc_parameters.csv'),
                                                 index=False)

        # Convert list of post hoc t test p value,
        # and f test p value in array
        # I squeeze it to get rid of the first (1, ...) dimension
        all_models_post_hoc_t_test_array = np.squeeze(np.array(all_network_post_hoc_test_p_values))
        all_models_f_test_p_values_array = np.array(all_network_f_test_p_values)

        # The user can choose different correction method
        p_values_correction_post_hoc_all_method = np.zeros((len(correction_method), len(models)*len(networks_list),
                                                            all_models_post_hoc_t_test_array.shape[1]))
        p_values_correction_f_test_all_method = np.zeros((len(correction_method), len(models)*len(networks_list)))

        # Loop over the correction method wanted by the user
        for correction in correction_method:

            # Correction with the chosen method for the p value of all between group t test
            post_hoc_t_test_corrected_p_values = multipletests(
                pvals=all_models_post_hoc_t_test_array.flatten(),
                method=correction,
                alpha=alpha)[1].reshape(all_models_post_hoc_t_test_array.shape)

            p_values_correction_post_hoc_all_method[correction_method.index(correction), ...] = \
                post_hoc_t_test_corrected_p_values

            # Correction with the chosen method for the p value of the whole model
            f_test_corrected_p_values = multipletests(
                pvals=all_models_f_test_p_values_array.flatten(),
                method=correction,
                alpha=alpha)[1].reshape(all_models_f_test_p_values_array.shape)

            p_values_correction_f_test_all_method[correction_method.index(correction), ...] =\
                f_test_corrected_p_values

        # Reshape the array of corrected p values for easy
        # writing in parameters files
        reshape_p_values_correction_post_hoc_all_method = p_values_correction_post_hoc_all_method.flatten().reshape(
            (len(correction_method), len(networks_list), len(models), all_models_post_hoc_t_test_array.shape[1]))

        reshape_p_values_correction_f_test_all_method = p_values_correction_f_test_all_method.flatten().reshape(
            (len(correction_method), len(networks_list), len(models), 1)
        )

        for model in models:
            for network in networks_list:

                regression_output_directory = folders_and_files_management.create_directory(
                    directory=os.path.join(root_analysis_directory, 'regression_analysis/ANOVA', kind,
                                           network))
                # Read the parameters file of the model for the f test first
                f_test_model_parameters = data_management.read_csv(os.path.join(
                    regression_output_directory, model + '_f_test_parameters.csv'))
                # Add the corrected p values for all the correction method
                post_hoc_model_parameters = data_management.read_csv(os.path.join(
                    regression_output_directory, model + '_post_hoc_parameters.csv'))

                # Append the corrected p values
                for correction in correction_method:
                    corrected_p_value_column_name = correction + '_p_value'
                    post_hoc_model_parameters[corrected_p_value_column_name] = \
                        reshape_p_values_correction_post_hoc_all_method[correction_method.index(correction),
                                                                        networks_list.index(network),
                                                                        models.index(model), ...]

                    f_test_model_parameters[corrected_p_value_column_name] = \
                        reshape_p_values_correction_f_test_all_method[correction_method.index(correction),
                                                                      networks_list.index(network),
                                                                      models.index(model), 0]

                    post_hoc_model_parameters.to_csv(os.path.join(regression_output_directory,
                                                                  model + '_post_hoc_parameters.csv'),
                                                     index=False)

                    f_test_model_parameters.to_csv(os.path.join(regression_output_directory,
                                                                model + '_f_test_parameters.csv'),
                                                   index=False)


def joint_models_correction(root_analysis_directory, kinds, models,
                            correction_methods,
                            networks=None, alpha=0.05):
    """Performs a joint models correction for the whole brains models,
    or the networks models.


    """
    for kind in kinds:
        if networks is None:
            # loop over the models
            all_models_p_values = []
            for model in models:
                # read current model dataframe
                model_df = pd.read_csv(os.path.join(root_analysis_directory, kind,
                                                    model + '_parameters.csv'))
                # Fetch the raw p_values
                all_models_p_values.append(model_df['p_value'])

            # Convert list to array of shape (n_models, n_tests)
            # each row (beta_0, beta_1, ...., beta_N) for the N variables in
            # the model.
            all_models_p_values_array = np.array(all_models_p_values)
            # Flatten the array
            all_models_p_values_array_flat = all_models_p_values_array.flatten()

            # Feed the flat p values array to chosen correction method
            for correction in correction_methods:
                _, corrected_p_values, _, _ = multipletests(pvals=all_models_p_values_array_flat,
                                                            alpha=alpha,
                                                            method=correction)

                # rebuild the array
                corrected_p_values_array = corrected_p_values.reshape(len(models), model_df.shape[0])

                # Append in the corresponding CSV model the corrected p values
                for model in models:
                    # read current model dataframe
                    model_df = pd.read_csv(os.path.join(root_analysis_directory, kind,
                                                        model + '_parameters.csv'))
                    # Append the corrected p value column
                    model_df[correction + "corrected_pvalues"] = corrected_p_values_array[models.index(model), ...]
                    # Overwrite the dataframe
                    model_df.to_csv(os.path.join(root_analysis_directory, kind,
                                                 model + '_parameters.csv'),
                                    index=False)

        else:
            all_models_p_values = []
            for model in models:
                network_model_p_values = []
                for network in networks:
                    model_network_df_path = Path(os.path.join(root_analysis_directory, kind, network,
                                                              model + '_parameters.csv'))
                    if model_network_df_path.exists():

                        model_network_df = pd.read_csv(os.path.join(root_analysis_directory, kind,
                                                                    network, model + '_parameters.csv'))
                        network_model_p_values.append(model_network_df['p_value'])

                    else:
                        pass
                # Convert to array
                network_model_p_values_array = np.array(network_model_p_values)

                # Append the network array of the model
                all_models_p_values.append(network_model_p_values_array)

            # convert it to array
            all_models_p_values_array = np.array(all_models_p_values)
            list_of_p_values = []

            for i in range(len(models)):
                list_of_p_values.append(all_models_p_values_array[i])
            # Flatten the nested list of p values
            all_models_p_values_list_flat = data_management.flatten(values=list_of_p_values)

            for correction in correction_methods:
                _, corrected_p_values, _, _ = multipletests(pvals=all_models_p_values_list_flat,
                                                            alpha=alpha,
                                                            method=correction)
                # Rebuild the nested list of p values
                corrected_p_values_unflatten = data_management.unflatten(corrected_p_values, list_of_p_values)

                # Write the corrected p values in corresponding CSV
                for model in models:
                    for network in networks:
                        model_network_df_path = Path(os.path.join(root_analysis_directory, kind, network,
                                                                  model + '_parameters.csv'))
                        if model_network_df_path.exists():
                            model_network_df = pd.read_csv(os.path.join(root_analysis_directory, kind,
                                                                        network, model + '_parameters.csv'))
                            # Append the corrected p value column
                            model_network_df[correction + "corrected_pvalues"] = \
                                corrected_p_values_unflatten[models.index(model)][networks.index(network), ...]

                            # Overwrite the dataframe
                            model_network_df.to_csv(os.path.join(root_analysis_directory, kind, network,
                                                                 model + '_parameters.csv'),
                                                    index=False)

                        else:
                            pass



# -*- coding: utf-8 -*-
"""
Created on Thu May 15 11:35:27 2014

@author:  edouard.duchesnay@cea.fr
@license: BSD-3-Clause
"""
import numpy as np
import mulm
import statsmodels.api as sm

n = 100
px = 5
py_info = 2
py_noize = 100

beta = np.array([1, 0, .5] + [0] * (px - 4) + [2]).reshape((px, 1))
X = np.hstack([np.random.randn(n, px-1), np.ones((n, 1))]) # X with intercept
Y = np.random.randn(n, py_info + py_noize)
# Causal model: add X on the first py_info variable
Y[:, :py_info] += np.dot(X, beta)

# t-test all the regressors (by default mulm and sm do two-tailed tests)
contrasts = np.identity(X.shape[1])

## OLS with statmodels, need to iterate over Y columns
sm_tvals = list()
sm_pvals = list()
for j in xrange(Y.shape[1]):
    mod = sm.OLS(Y[:, j], X)
    sm_ttest = mod.fit().t_test(contrasts)
    sm_tvals.append(sm_ttest.tvalue)
    sm_pvals.append(sm_ttest.pvalue)

sm_tvals = np.asarray(sm_tvals).T
sm_pvals = np.asarray(sm_pvals).T

## OLS with MULM
mod = mulm.MUOLS(Y, X)
mulm_tvals, mulm_pvals, mulm_df = mod.fit().t_test(contrasts, pval=True, two_tailed=True)

# Check that results ar similar
np.allclose(mulm_tvals, sm_tvals)
np.allclose(mulm_pvals, sm_pvals)
